# ANDES, a power system simulation tool for research.
#
# Copyright 2015-2017 Hantao Cui
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base class for building ANDES models
"""

import importlib
import logging
import sys
from logging import DEBUG, INFO, WARNING, ERROR

import numpy as np
from cvxopt import matrix
from cvxopt import mul, div

from ..utils.math import agtb, altb
from ..utils.tab import Tab

logger = logging.getLogger(__name__)

pd = None


class ModelBase(object):
    """base class for power system device models"""

    def __init__(self, system, name):
        """meta-data to be overloaded by subclasses"""

        # metadata list and dictionaries
        self._param_attr_lists = ('_params', '_zeros', '_mandatory', '_powers',
                                  '_voltages', '_currents', '_z', '_y', '_r',
                                  '_g', '_dccurrents', '_dcvoltages', '_times')

        self._param_attr_dicts = ('_data', '_units', '_descr')

        self.system = system
        # self.n = 0    # device count
        self.u = []  # device status
        self.idx = []  # external index list
        self.uid = {}  # mapping from `self.idx` to unique positional indices
        self.name = []  # element name list

        # identifications
        self._name = name
        self._group = None
        self._category = None

        # interfaces
        self._ac = {}  # ac bus variables
        self._dc = {}  # dc node variables
        self._ctrl = {}  # controller interfaces

        # variables
        self._states = []
        self._algebs = []
        self._algebs_intf = []

        self._states_descr = {}
        self._algebs_descr = {}

        # variable names
        self._unamex = []
        self._unamey = []
        self._fnamex = []
        self._fnamey = []

        # `self.mdl_from`:
        #    a list of lists, which contain tuples of (model, idx)
        #    of elements connected to the corresponding element
        # `self.mdl_to`:
        #    a list of lists, which contain tuples of (model, idx) of
        #    elements the corresponding element connects to

        self.mdl_from = []
        self.mdl_to = []

        # parameters to be converted to matrix
        self._params = ['u', 'Sn', 'Vn']

        # default parameter data
        self._data = {
            'u': 1,
            'Sn': 100.0,
            'Vn': 110.0,
        }

        # units of parameters
        self._units = {
            'u': 'bool',
            'Sn': 'MVA',
            'Vn': 'kV',
        }

        # variable descriptions
        self._descr = {
            'u': 'Connection status',
            'Sn': 'Power rating',
            'Vn': 'AC voltage rating',
        }
        # non-zero parameters
        self._zeros = ['Sn', 'Vn']

        # mandatory variables
        self._mandatory = []

        # service/temporary variables
        self._service = []
        self._service_ty = []

        self._store = {}
        self._snapshot = {}

        # equation storage
        self._equations = []

        # parameters to be per-unitized
        self._powers = []  # powers, inertia and damping
        self._voltages = []  # ac voltages
        self._currents = []  # ac currents
        self._z = []  # ac impedance
        self._y = []  # ac admittance

        self._dccurrents = []  # dc currents
        self._dcvoltages = []  # dc voltages
        self._r = []  # dc resistance
        self._g = []  # dc susceptance

        self._times = []  # time constants

        self._event_times = []  # event occurrance times

        # property functions this device has

        self.calls = dict(
            pflow=False,
            addr1=False,
            init0=False,
            init1=False,
            jac0=False,
            windup=False,
            gcall=False,
            fcall=False,
            gycall=False,
            fxcall=False,
            series=False,
            shunt=False,
            flows=False,
            dcseries=False,
        )

        self._flags = {
            'sysbase': False,
            'allocate': False,
            'address': False,
        }

        self._config = {
            'address_group_by': 'element',
            'is_series': False,
        }

        # pandas.DataFrame
        self.param_dict = {}
        self.param_df = {}

        self.define()

    def define(self):
        """
        Hook function where derived models define parameters, variables,
        service constants, and equations

        :return:
        """
        # raise NotImplemented('Subclasses must overwrite this method')
        pass

    def _init(self):
        """
        Convert model metadata to class attributes.

        This function is called automatically after ``define()`` in new
        versions.

        :return: None
        """
        if not self._name:
            raise ValueError("Must specify a name for class {}".format(self.__class__.__name__))
        if not self._group:
            raise ValueError("Must specify a group name for {}".format(self._name))

        # self.n = 0
        self.u = []
        self.name = []
        self.idx = []
        self.uid = {}

        if (not self._unamey) or (len(self._unamey) != len(self._algebs)):  # allow multiple calls to `_init`
            self._unamey = self._algebs

        if (not self._unamex) or (len(self._unamex) != len(self._states)):  # allow multiple calls
            self._unamex = self._states

        for item in self._data.keys():
            self.__dict__[item] = []

        for bus in self._ac.keys():
            for var in self._ac[bus]:
                self.__dict__[var] = []
        for node in self._dc.keys():
            for var in self._dc[node]:
                self.__dict__[var] = []

        for var in self._states + self._algebs + self._service:
            self.__dict__[var] = []

        self._flags['sysbase'] = False
        self._flags['allocate'] = False
        self._flags['address'] = False

    def param_define(self, param, default, unit='', descr='', tomatrix=True, nonzero=False, mandatory=False,
                     power=False, voltage=False, current=False, z=False, y=False, r=False, g=False,
                     dccurrent=False, dcvoltage=False, time=False, event_time=False, **kwargs):
        """
        Define a parameter in the model

        :param tomatrix: convert this parameter list to matrix
        :param param: parameter name
        :param default: parameter default value
        :param unit: parameter unit
        :param descr: description
        :param nonzero: is non-zero
        :param mandatory: is mandatory
        :param power: is a power value in the `self.Sn` base
        :param voltage: is a voltage value in the `self.Vn` base
        :param current: is a current value in the device base
        :param z: is an impedance value in the device base
        :param y: is an admittance value in the device base
        :param r: is a dc resistance value in the device base
        :param g: is a dc conductance value in the device base
        :param dccurrent: is a dc current value in the device base
        :param dcvoltage: is a dc votlage value in the device base
        :param time: is a time value in the device base
        :param event_time: is a variable for timed event

        :type param: str
        :type tomatrix: bool
        :type default: str, float
        :type unit: str
        :type descr: str
        :type nonzero: bool
        :type mandatory: bool
        :type power: bool
        :type voltage: bool
        :type current: bool
        :type z: bool
        :type y: bool
        :type r: bool
        :type g: bool
        :type dccurrent: bool
        :type dcvoltage: bool
        :type time: bool
        :type event_time: bool
        """
        if param in self._data:
            raise KeyError('Param <{}> already defined in _data'.format(param))
        if param in self._algebs:
            raise KeyError('Param <{}> already defined in _algebs'.format(param))
        if param in self._states:
            raise KeyError('Param <{}> already defined in _states'.format(param))
        if param in self._service:
            raise KeyError('Param <{}> already defined in _service'.format(param))

        self._data.update({param: default})
        if unit:
            self._units.update({param: unit})
        if descr:
            self._descr.update({param: descr})
        if tomatrix:
            self._params.append(param)
        if nonzero:
            self._zeros.append(param)
        if mandatory:
            self._mandatory.append(param)
        if power:
            self._powers.append(param)
        if voltage:
            self._voltages.append(param)
        if current:
            self._currents.append(param)
        if z:
            self._z.append(param)
        if y:
            self._y.append(param)
        if r:
            self._r.append(param)
        if g:
            self._g.append(param)
        if dccurrent:
            self._dccurrents.append(param)
        if dcvoltage:
            self._dcvoltages.append(param)
        if time:
            self._times.append(param)
        if event_time:
            self._event_times.append(param)

    def var_define(self, variable, ty, fname, descr='', uname=''):
        """
        Define a variable in the model

        :param fname: LaTex formatted variable name string
        :param uname: unformatted variable name string, `variable` as default
        :param variable: variable name
        :param ty: type code in ``('x', 'y')``
        :param descr: variable description

        :type variable: str
        :type ty: str
        :type descr: str
        :return:
        """
        if ty not in ('x', 'y'):
            raise ValueError('Type <{}> of the variable <{}> is invalid'.format(ty, variable))

        if not uname:
            uname = variable

        if ty == 'x':
            self._states.append(variable)
            self._fnamex.append(fname)
            self._unamex.append(uname)
            if descr:
                self._states_descr.update({variable: descr})
        elif ty == 'y':
            self._algebs.append(variable)
            self._fnamey.append(fname)
            self._unamey.append(uname)
            if descr:
                self._algebs_descr.update({variable: descr})

    def service_define(self, service, ty):
        """
        Add a service variable of type ``ty`` to this model

        :param str service: variable name
        :param type ty: variable type
        :return: None
        """

        if service in self._data:
            raise NameError('Service variable <{}> already defined in _data'.format(service))
        if service in self._algebs:
            raise NameError('Service variable <{}> already defined in _algebs'.format(service))
        if service in self._states:
            raise NameError('Service variable <{}> already defined in _states'.format(service))

        self._service.append(service)
        self._service_ty.append(ty)

    def get_uid(self, idx):
        """
        Return the `uid` of the elements with the given `idx`

        :param list, matrix idx: external indices
        :type idx: list, matrix
        :return: a matrix of uid
        """
        if idx is None:
            raise IndexError('get_uid cannot take None as idx')

        if isinstance(idx, (int, float, str)):
            return self.uid[idx]

        ret = []
        for i in idx:
            tmp = self.uid.get(i, None)
            if tmp is None:
                raise IndexError('Model <{}> does not contain any element with idx={}. '
                                 'Check the definitions in your case file.'.format(self._name, i))

            ret.append(self.uid[i])

        return ret

    def get_idx(self, uid):
        """
        Return the ``idx`` of the elements whose internal indices are ``uid``

        :param uid: uid of elements to query
        :return: idx of the elements
        """
        return [self.idx[i] for i in uid]

    def get_field(self, field, idx=None, astype=None):
        """
        Return `self.field` for the elements labeled by `idx`

        :param astype: type cast of the return value
        :param field: field name of this model
        :param idx: element indices, will be the whole list if not specified
        :return: field values
        """
        if astype not in (None, list, matrix):
            raise TypeError("Unsupported return type {}".format(astype))

        ret = None

        if idx is None:
            idx = self.idx

        # ===================disable warning ==============================
        # if field in self._service:
        # logger.warning(
        #     'Reading service variable <{model}.{field}> could be unsafe.'
        #         .format(field=field, model=self._name)
        # )
        # =================================================================

        uid = self.get_uid(idx)

        field_data = self.__dict__[field]

        if isinstance(field_data, matrix):
            ret = field_data[uid]
        elif isinstance(field_data, list):
            if isinstance(idx, (float, int, str)):
                ret = field_data[uid]
            else:
                ret = [field_data[x] for x in uid]

        if astype is not None:
            ret = astype(ret)

        return ret

    def set_field(self, field, idx, value, sysbase=False):
        """
        Set the field of an element to the given value

        Parameters
        ----------
        field
        idx
        value
        sysbase

        Returns
        -------

        """
        if field not in self.__dict__:
            logger.error('Field <{}> does not exist.'.format(field))
            return

        if idx not in self.idx:
            logger.error('Idx <{}> is not valid for model <{}>'.format(idx, self._name))
            logger.error('Available idx are: {}'.format(self.idx))
            raise IndexError

        uid = self.get_uid(idx)
        old_type = type(self.__dict__[field][uid])

        in_store = True if field in self._store else False

        if (not sysbase) and in_store:
            self._store[field][uid] = old_type(value)
            logger.info('{}: set field <{}> of <{}> to {} in _store'.format(self._name, field, idx, value))
        else:
            self.__dict__[field][uid] = old_type(value)
            logger.info('{}: set field <{}> of <{}> to {} in class dict'.format(self._name, field, idx, value))

    def _alloc(self):
        """
        Allocate empty memory for dae variable indices.
        Called in device setup phase.

        :return: None
        """
        nzeros = [0] * self.n
        for var in self._states:
            self.__dict__[var] = nzeros[:]
        for var in self._algebs:
            self.__dict__[var] = nzeros[:]

    def data_to_dict(self, sysbase=False):
        """
        Return the loaded model parameters as one dictionary.

        Each key of the dictionary is a parameter name, and the value is a
        list of all the parameter values.

        :param sysbase: use system base quantities
        :type sysbase: bool
        """
        if not isinstance(sysbase, bool):
            raise TypeError('Argument <sysbase> must be True or False')

        ret = {}

        for key in self.data_keys:
            if (not sysbase) and (key in self._store):
                val = self._store[key]
            else:
                val = self.__dict__[key]

            ret[key] = val

        return ret

    @property
    def n(self):
        """
        Return the count of elements

        Returns
        -------
        int:
            count of elements
        """

        return len(self.idx)

    @property
    def data_keys(self):
        """
        Returns a list of all parameters plus ``name`` and ``idx``

        Returns
        -------
        list of parameters
        """

        keys = ['idx', 'name', 'u'] + sorted(self._data.keys())
        return set(keys)

    def fxcall_idx(self):
        """
        Precompute the row and column indices of the non-zero elements in the df/dx matrix

        Returns
        -------
        (row indices, col indices)
        """

        pass

    def data_to_list(self, sysbase=False):
        """
        Return the loaded model data as a list of dictionaries.

        Each dictionary contains the full parameters of an element.
        :param sysbase: use system base quantities
        :type sysbase: bool
        """
        ret = list()

        # for each element
        for i in range(self.n):
            # read the parameter values and put in the temp dict ``e``
            e = {}
            for key in self.data_keys:
                if sysbase and (key in self._store):
                    val = self._store[key][i]
                else:
                    val = self.__dict__[key][i]
                e[key] = val

            ret.append(e)

        return ret

    def data_from_list(self, data):
        """
        Populate model parameters from a list of parameter dictionaries

        Parameters
        ----------
        data : list
            List of parameter dictionaries

        Returns
        -------
        None
        """
        for item in data:
            self.elem_add(**item)

    def data_from_dict(self, data):
        """
        Populate model parameters from a dictionary of parameters

        Parameters
        ----------
        data : dict
            List of parameter dictionaries

        Returns
        -------
        None
        """

        nvars = []
        for key, val in data.items():
            self.__dict__[key].extend(val)

            # assure the same parameter matrix size
            if len(nvars) > 1 and len(val) != nvars[-1]:
                raise IndexError(
                    'Model <{}> parameter <{}> must have the same length'.
                    format(self._name, key))
            nvars.append(len(val))

        # assign idx-uid mapping
        for i, idx in zip(range(self.n), self.idx):
            self.uid[idx] = i

    def data_to_df(self, sysbase=False):
        """
        Return a pandas.DataFrame of device parameters.
        :param sysbase: save per unit values in system base
        """

        p_dict_comp = self.data_to_dict(sysbase=sysbase)
        if self._check_pd():
            self.param_df = pd.DataFrame(data=p_dict_comp).set_index('idx')
            return self.param_df

    def var_to_df(self):
        """
        Return the current var_to_df of variables

        :return: pandas.DataFrame
        """
        ret = {}
        self._check_pd()

        if self._flags['address'] is False:
            return pd.DataFrame.from_dict(ret)

        ret.update({'name': self.name})
        ret.update({'idx': self.idx})

        for x in self._states:
            idx = self.__dict__[x]
            ret.update({x: self.system.dae.x[idx]})
        for y in self._algebs:
            idx = self.__dict__[y]
            ret.update({y: self.system.dae.y[idx]})

        var_df = pd.DataFrame.from_dict(ret).set_index('idx')

        return var_df

    @staticmethod
    def _check_pd():
        """
        Import pandas to globals() if not exist
        """
        if globals()['pd'] is None:
            try:
                globals()['pd'] = importlib.import_module('pandas')
            except ImportError:
                logger.warning("Pandas import error.")
                return False

        return True

    def param_remove(self, param: 'str') -> None:
        """
        Remove a param from this model

        :param param: name of the parameter to be removed
        :type param: str
        """
        for attr in self._param_attr_dicts:
            if param in self.__dict__[attr]:
                self.__dict__[attr].pop(param)

        for attr in self._param_attr_lists:
            if param in self.__dict__[attr]:
                self.__dict__[attr].remove(param)

    def param_alter(self, param, default=None, unit=None, descr=None, tomatrix=None, nonzero=None, mandatory=None,
                    power=None, voltage=None, current=None, z=None, y=None, r=None, g=None, dccurrent=None,
                    dcvoltage=None, time=None, **kwargs):
        """
        Set attribute of an existing parameter.
        To be used to alter an attribute inherited from parent models.

        See .. self.param_define for argument descriptions.
        """
        if param not in self._data:
            logger.error('Param <{}> does not exist in {}._data. '
                         'param_alter cannot continue'.format(param, self.name))
            return False

        def alter_attr(p, attr, value):
            """Set self.__dict__[attr] for param based on value
            """
            if value is None:
                return
            elif (value is True) and (p not in self.__dict__[attr]):
                self.__dict__[attr].append(p)
            elif (value is False) and (p in self.__dict__[attr]):
                self.__dict__[attr].remove(p)
            else:
                self.log('No need to alter {} for {}'.format(attr, p))

        if default is not None:
            self._data.update({param: default})
        if unit is not None:
            self._units.update({param: unit})
        if descr is not None:
            self._descr.update({param: descr})

        alter_attr(param, '_params', tomatrix)
        alter_attr(param, '_zeros', nonzero)
        alter_attr(param, '_mandatory', mandatory)
        alter_attr(param, '_powers', power)
        alter_attr(param, '_voltages', voltage)
        alter_attr(param, '_currents', current)
        alter_attr(param, '_z', z)
        alter_attr(param, '_y', y)
        alter_attr(param, '_r', r)
        alter_attr(param, '_g', g)
        alter_attr(param, '_dcvoltages', dcvoltage)
        alter_attr(param, '_dccurrents', dccurrent)
        alter_attr(param, '_times', time)

    def eq_add(self, expr, var, intf=False):
        """
        Add an equation to this model.

        An equation is associated with the addresses of a variable.  The
        number of equations must equal that of variables.

        Stored to ``self._equations`` is a tuple of ``(expr, var, intf,
        ty)`` where ``ty`` is in ('f', 'g')

        :param str expr: equation expression
        :param str var: variable name to be associated with
        :param bool intf: if this equation is added to an interface equation,
            namely, an equation outside this model

        :return: None

        """

        # determine the type of the equation, ('f', 'g') based on var

        # We assume that all differential equations are only modifiable by
        #   the model itself
        # Only interface to algebraic varbailes of external models

        ty = ''
        if var in self._algebs:
            ty = 'g'
        elif var in self._states:
            ty = 'f'
        else:
            for _, item in self._ac:
                if var in item:
                    ty = 'g'
                    if intf is False:
                        intf = True
            for _, item in self._dc:
                if var == item:
                    ty = 'g'
                    if intf is False:
                        intf = True
        if ty == '':
            self.log(
                'Equation associated with interface variable {var} '
                'assumed as algeb'.
                format(var=var),
                DEBUG)
            ty = 'g'

        self._equations.append((expr, var, intf, ty))

    def read_data_ext(self, model: str, field: str, idx=None, astype=None):
        """
        Return a field of a model or group at the given indices.

        The default return type if `cvxopt.matrix`. If returning a list is preferred,
        use `astype=list`.

        :param str model: name of the group or model to retrieve
        :param str field: name of the field
        :param list, int, float str idx: idx of elements to access
        :param type astype: type cast
        :return:
        """
        ret = list()

        if model in self.system.devman.devices:
            ret = self.system.__dict__[model].get_field(field, idx)

        elif model in self.system.devman.group.keys():
            # ===============================================================
            # Since ``self.system.devman.group`` is an unordered dictionary,
            #   ``idx`` must be given to retrieve ``field`` across models.
            #
            # Returns a matrix by default
            # ===============================================================
            if idx is None:
                raise IndexError('idx must be provided when accessing <{}> group field'.format(model))

            astype = matrix

            for item in idx:
                dev_name = self.system.devman.group[model].get(item, None)
                if dev_name is None:
                    raise KeyError('Group <{}> does not contain element whose idx={}'.format(model, item))
                ret.append(self.read_data_ext(dev_name, field, idx=item))

        else:
            raise NameError('Model or Group <{0}> does not exist.'.format(model))

        if (ret is None) or isinstance(ret, (int, float, str)):
            return ret
        elif astype is None:
            return ret
        else:
            return astype(ret)

    def copy_data_ext(self, model, field, dest=None, idx=None, astype=None):
        """
        Retrieve the field of another model and store it as a field.

        :param model: name of the source model being a model name or a group name
        :param field: name of the field to retrieve
        :param dest: name of the destination field in ``self``
        :param idx: idx of elements to access
        :param astype: type cast

        :type model: str
        :type field: str
        :type dest: str
        :type idx: list, matrix
        :type astype: None, list, matrix

        :return: None

        """
        # use default destination

        if not dest:
            dest = field
        if dest in self._states + self._algebs:
            raise KeyError('Destination field <{}> already exist in <{}>. '
                           'copy_data_ext cannot continue'.format(field, model))

        self.__dict__[dest] = self.read_data_ext(
            model, field, idx, astype=astype)

        if idx is not None:
            if len(idx) == self.n:
                self.link_to(model, idx, self.idx)

    def elem_add(self, idx=None, name=None, **kwargs):
        """
        Add an element of this model

        Parameters
        ----------
        idx : str, int or float, optional
            The unique external reference to this element.
        name : str, optional
            The name of this element. Automatically assigned if not provided.

        kwargs
            A dictionary containing the (field, value) pairs of the element parameters

        Returns
        -------
        str or int
            The allocated `idx` of the added element
        """

        idx = self.system.devman.register_element(dev_name=self._name, idx=idx)
        self.system.__dict__[self._group].register_element(self._name, idx)

        self.uid[idx] = self.n
        self.idx.append(idx)

        self.mdl_to.append(list())
        self.mdl_from.append(list())

        # self.n += 1

        if name is None:
            self.name.append(self._name + ' ' + str(self.n))
        else:
            self.name.append(name)

        # check mandatory parameters
        for key in self._mandatory:
            if key not in kwargs.keys():
                self.log(
                    'Mandatory parameter <{:s}.{:s}> missing'.format(
                        self.name[-1], key), ERROR)
                sys.exit(1)

        # set default values
        for key, value in self._data.items():
            self.__dict__[key].append(value)

        # overwrite custom values
        for key, value in kwargs.items():
            if key not in self._data:
                self.log(
                    'Parameter <{:s}.{:s}> is not used.'.format(
                        self.name[-1], key), WARNING)
                continue
            self.__dict__[key][-1] = value

            # check data consistency
            if not value and key in self._zeros:
                if key == 'Sn':
                    default = self.system.mva
                elif key == 'fn':
                    default = self.system.config.freq
                else:
                    default = self._data[key]
                self.__dict__[key][-1] = default
                logger.debug('Using default value for <{:s}.{:s}>'.format(
                    self.name[-1], key))

        return idx

    def elem_remove(self, idx=None):
        """
        Remove elements labeled by idx from this model instance.

        :param list,matrix idx: indices of elements to be removed
        :return: None
        """
        if idx is not None:
            if idx in self.uid:
                key = idx
                item = self.uid[idx]
            else:
                self.log('The item <{:s}> does not exist.'.format(idx), ERROR)
                return None
        else:
            return None

        convert = False
        if isinstance(self.__dict__[self._params[0]], matrix):
            self._param_to_list()
            convert = True

        # self.n -= 1
        self.uid.pop(key, '')
        self.idx.pop(item)
        self.mdl_from.pop(item)
        self.mdl_to.pop(item)

        for x, y in self.uid.items():
            if y > item:
                self.uid[x] = y - 1

        for param in self._data:
            self.__dict__[param].pop(item)

        for param in self._service:
            if len(self.__dict__[param]) == (self.n + 1):
                if isinstance(self.__dict__[param], list):
                    self.__dict__[param].pop(item)
                elif isinstance(self.__dict__[param], matrix):
                    service = list(self.__dict__[param])
                    service.pop(item)
                    self.__dict__[param] = matrix(service)

        for x in self._states:
            if len(self.__dict__[x]):
                self.__dict__[x].pop(item)

        for y in self._algebs:
            if self.__dict__[y]:
                self.__dict__[y].pop(item)

        for key, param in self._ac.items():
            if isinstance(param, list):
                for subparam in param:
                    if len(self.__dict__[subparam]):
                        self.__dict__[subparam].pop(item)
            else:
                self.__dict__[param].pop(item)

        for key, param in self._dc.items():
            self.__dict__[param].pop(item)

        self.name.pop(item)
        if convert and self.n:
            self._param_to_matrix()

    def reload_new_param(self):
        """
        Reload the new params modified in `self._store` and then convert to sysbase and store in class.

        Returns
        -------

        """
        if not self._flags['sysbase']:  # parameters have not been converted to sys base
            self.data_to_sys_base()
        else:  # parameters are in sys base. `self._store` has the parameters in elem base

            for key, val in self._store:
                self.__dict__[key] = val
            self._flags['sysbase'] = False
            self.data_to_sys_base()
            logger.debug('Reloaded system parameter for _store.')

    def data_to_sys_base(self):
        """
        Converts parameters to system base. Stores a copy in ``self._store``.
        Sets the flag ``self.flag['sysbase']`` to True.

        :return: None
        """
        if (not self.n):
            return
        if self._flags['sysbase']:
            logger.debug('Model <{}> is not in sysbase. data_to_elem_base() aborted.'.format(self._name))
            return

        Sb = self.system.mva
        Vb = matrix([])
        if 'bus' in self._ac.keys():
            Vb = self.read_data_ext('Bus', 'Vn', idx=self.bus)
        elif 'bus1' in self._ac.keys():
            Vb = self.read_data_ext('Bus', 'Vn', idx=self.bus1)

        for var in self._voltages:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Vn)
            self.__dict__[var] = div(self.__dict__[var], Vb)
        for var in self._powers:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Sn)
            self.__dict__[var] /= Sb
        for var in self._currents:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Sn)
            self.__dict__[var] = div(self.__dict__[var], self.Vn)
            self.__dict__[var] = mul(self.__dict__[var], Vb)
            self.__dict__[var] /= Sb
        if len(self._z) or len(self._y):
            Zn = div(self.Vn**2, self.Sn)
            Zb = (Vb**2) / Sb
            for var in self._z:
                self._store[var] = self.__dict__[var]
                self.__dict__[var] = mul(self.__dict__[var], Zn)
                self.__dict__[var] = div(self.__dict__[var], Zb)
            for var in self._y:
                self._store[var] = self.__dict__[var]
                if self.__dict__[var].typecode == 'd':
                    self.__dict__[var] = div(self.__dict__[var], Zn)
                    self.__dict__[var] = mul(self.__dict__[var], Zb)
                elif self.__dict__[var].typecode == 'z':
                    self.__dict__[var] = div(self.__dict__[var], Zn + 0j)
                    self.__dict__[var] = mul(self.__dict__[var], Zb + 0j)

        if len(self._dcvoltages) or len(self._dccurrents) or len(
                self._r) or len(self._g):
            dckey = sorted(self._dc.keys())[0]
            Vbdc = self.read_data_ext('Node', 'Vdcn', self.__dict__[dckey])
            Ib = div(Sb, Vbdc)
            Rb = div(Vbdc, Ib)

        for var in self._dcvoltages:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Vdcn)
            self.__dict__[var] = div(self.__dict__[var], Vbdc)

        for var in self._dccurrents:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], self.Idcn)
            self.__dict__[var] = div(self.__dict__[var], Ib)

        for var in self._r:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = div(self.__dict__[var], Rb)

        for var in self._g:
            self._store[var] = self.__dict__[var]
            self.__dict__[var] = mul(self.__dict__[var], Rb)

        self._flags['sysbase'] = True

    def data_to_elem_base(self):
        """
        Convert parameter data to element base

        Returns
        -------
        None
        """
        if not self.n:
            return
        if self._flags['sysbase'] is False:
            logger.debug('Model <{}> is not in sysbase. data_to_elem_base() aborted.'.format(self._name))
            return

        for key, val in self._store.items():
            self.__dict__[key] = val

        self._flags['sysbase'] = False

    def setup(self):
        """
        Set up empty class structure and allocate empty memory
        for variable addresses.

        :return: None
        """
        self._param_to_matrix()  # fixes an issue when PV.n=0
        self._alloc()

    def _intf_network(self):
        """
        Retrieve the ac and dc network interface variable indices.

        :Example:

            ``self._ac = {'bus1': (a1, v1)}`` gives
             - indices: self.bus1
             - system.Bus.a -> self.a1
             - system.Bus.v -> self.v1

            ``self._dc = {'node1': v1}`` gives
              - indices: self.node1
              - system.Node.v-> self.v1

        :return: None
        """

        for key, val in self._ac.items():
            self.copy_data_ext(
                model='Bus', field='a', dest=val[0], idx=self.__dict__[key])
            self.copy_data_ext(
                model='Bus', field='v', dest=val[1], idx=self.__dict__[key])

        for key, val in self._dc.items():
            self.copy_data_ext(
                model='Node', field='v', dest=val, idx=self.__dict__[key])

        # check for interface voltage differences
        self._check_Vn()

    def _intf_ctrl(self):
        """
        Retrieve variable indices of controlled models.

        Control interfaces are specified in ``self._ctrl``.
        Each ``key:value`` pair has ``key`` being the variable names
        for the reference idx and ``value`` being a tuple of
        ``(model name, field to read, destination field, return type)``.

        :Example:

            ``self._ctrl = {'syn', ('Synchronous', 'omega', 'w', list)}``
             - indices: self.syn
             - self.w = list(system.Synchronous.omega)

        :return: None
        """

        for key, val in self._ctrl.items():
            model, field, dest, astype = val
            self.copy_data_ext(
                model, field, dest=dest, idx=self.__dict__[key], astype=astype)

    def _addr(self):
        """
        Assign dae addresses for algebraic and state variables.

        Addresses are stored in ``self.__dict__[var]``.
        ``dae.m`` and ``dae.n`` are updated accordingly.

        Returns
        -------
        None

        """
        group_by = self._config['address_group_by']

        if group_by not in ('element', 'variable'):
            raise KeyError('Argument group_by value <{}> is not element or variable'.format(group_by))

        if self._flags['address'] is True:
            logger.debug('{} addresses exist. Skipped.'.format(self._name))
            return

        m0 = self.system.dae.m
        n0 = self.system.dae.n
        mend = m0 + len(self._algebs) * self.n
        nend = n0 + len(self._states) * self.n

        if group_by == 'variable':
            for idx, item in enumerate(self._algebs):
                self.__dict__[item] = list(
                    range(m0 + idx * self.n, m0 + (idx + 1) * self.n))
            for idx, item in enumerate(self._states):
                self.__dict__[item] = list(
                    range(n0 + idx * self.n, n0 + (idx + 1) * self.n))
        elif group_by == 'element':
            for idx, item in enumerate(self._algebs):
                self.__dict__[item] = list(
                    range(m0 + idx, mend, len(self._algebs)))
            for idx, item in enumerate(self._states):
                self.__dict__[item] = list(
                    range(n0 + idx, nend, len(self._states)))

        self.system.dae.m = mend
        self.system.dae.n = nend

        self._flags['address'] = True

    def _varname(self):
        """
        Set up variable names in ``self.system.varname``.

        Variable names follows the convention ``VariableName,Model Name``.
        A maximum of 24 characters are allowed for each variable.

        :return: None
        """
        if not self._flags['address']:
            self.log('Unable to assign Varname before allocating address',
                     ERROR)
            return

        varname = self.system.varname
        for i in range(self.n):
            iname = str(self.name[i])

            for e, var in enumerate(self._states):
                unamex = self._unamex[e]
                fnamex = self._fnamex[e]
                idx = self.__dict__[var][i]

                varname.unamex[idx] = '{} {}'.format(unamex, iname)[:33]
                varname.fnamex[idx] = '$' + r'{}\ {}'.format(
                    fnamex, iname.replace(' ', r'\ '))[:33] + '$'

            for e, var in enumerate(self._algebs):
                unamey = self._unamey[e]
                fnamey = self._fnamey[e]
                idx = self.__dict__[var][i]

                varname.unamey[idx] = '{} {}'.format(unamey, iname)[:33]
                varname.fnamey[idx] = '$' + r'{}\ {}'.format(
                    fnamey, iname.replace(' ', r'\ '))[:33] + '$'

    def _param_to_matrix(self):
        """
        Convert parameters defined in `self._params` to `cvxopt.matrix`

        :return None
        """
        for item in self._params:
            self.__dict__[item] = matrix(self.__dict__[item], tc='d')

    def _param_to_list(self):
        """
        Convert parameters defined in `self._param` to list

        :return None
        """
        for item in self._params:
            self.__dict__[item] = list(self.__dict__[item])

    def log(self, msg, level=INFO):
        """Record a line of log in logger

        :param str msg: content of the messag

        :param level: logging level
        :return: None

        """
        logger.log(level, '<{}> - '.format(self._name) + msg)

    def init_limit(self, key, lower=None, upper=None, limit=False):
        """ check if data is within limits. reset if violates"""
        above = agtb(self.__dict__[key], upper)
        for idx, item in enumerate(above):
            if item == 0.:
                continue
            maxval = upper[idx]
            self.log(
                '{0} <{1}.{2}> above its maximum of {3}.'.format(
                    self.name[idx], self._name, key, maxval), ERROR)
            if limit:
                self.__dict__[key][idx] = maxval

        below = altb(self.__dict__[key], lower)
        for idx, item in enumerate(below):
            if item == 0.:
                continue
            minval = lower[idx]
            self.log(
                '{0} <{1}.{2}> below its minimum of {3}.'.format(
                    self.name[idx], self._name, key, minval), ERROR)
            if limit:
                self.__dict__[key][idx] = minval

    def __repr__(self):

        ret = ''
        ret += '\n'
        ret += 'Model <{:s}> parameters in element base\n'.format(self._name)
        ret += self.data_to_df(sysbase=False).to_string()

        ret += '\n'
        ret += 'Model <{:s}> variable snapshot\n'.format(self._name)
        ret += self.var_to_df().to_string()

        return ret

    def __str__(self):
        super(ModelBase, self).__repr__()
        return '<{name}> with {n} elements [{mem}]'.format(
            name=self._name,
            mem=hex(id(self)),
            n=self.n,
        )

    def doc(self, export='plain'):
        """
        Build help document into a table

        :param ('plain', 'latex') export: export format
        :param save: save to file ``help_model.extension`` or not
        :param writemode: file write mode

        :return: None
        """
        title = '<{}.{}>'.format(self._group, self._name)
        table = Tab(export=export, title=title, descr=self.__doc__)
        rows = []
        keys = sorted(self._data.keys())
        for key in keys:
            val = self._data[key]
            suf = ''
            if key in self._mandatory:
                suf = ' *'

            elif key in self._powers + \
                    self._voltages + \
                    self._currents + \
                    self._z + \
                    self._y + \
                    self._dccurrents + \
                    self._dcvoltages + \
                    self._r + \
                    self._g + \
                    self._times:
                suf = ' #'

            c1 = key + suf
            c2 = self._descr.get(key, '')
            c3 = val
            c4 = self._units.get(key, '-')
            rows.append([c1, c2, c3, c4])

        table.add_rows(rows, header=False)
        table.header(['Parameter', 'Description', 'Default', 'Unit'])

        if export == 'plain':
            pass
        elif export == 'latex':
            raise NotImplementedError('LaTex output not implemented')

        return table.draw()

    def check_limit(self, varname, vmin=None, vmax=None):
        """
        Check if the variable values are within the limits.

        Return False if fails.

        """
        retval = True

        if varname not in self.__dict__:
            logger.error('Variable <{}> invalid. check_limit cannot continue.'.format(varname))

        if varname in self._algebs:
            val = self.system.dae.y[self.__dict__[varname]]
        elif varname in self._states:
            val = self.system.dae.x[self.__dict__[varname]]
        else:  # service or temporary variable
            val = matrix(self.__dict__[varname])

        vmin = matrix(self.__dict__[vmin])
        comp = altb(val, vmin)
        comp = mul(self.u, comp)
        for c, n, idx in zip(comp, self.name, range(self.n)):
            if c == 1:
                v = val[idx]
                vm = vmin[idx]
                self.log(
                    'Init of <{}.{}>={:.4g} is lower than min={:6.4g}'.format(
                        n, varname, v, vm), ERROR)
                retval = False

        vmax = matrix(self.__dict__[vmax])
        comp = agtb(val, vmax)
        comp = mul(self.u, comp)
        for c, n, idx in zip(comp, self.name, range(self.n)):
            if c == 1:
                v = val[idx]
                vm = vmax[idx]
                self.log(
                    'Init of <{}.{}>={:.4g} is higher than max={:.4g}'.format(
                        n, varname, v, vm), ERROR)
                retval = False
        return retval

    def on_bus(self, bus_idx):
        """
        Return the indices of elements on the given buses for shunt-connected
        elements

        :param bus_idx: idx of the buses to which the elements are connected
        :return: idx of elements connected to bus_idx

        """
        if not hasattr(self, 'bus'):
            logger.warning('Model <{}> does not have the bus field. '
                           'on_bus() cannot continue.'.format(self._name))

        ret = []
        if isinstance(bus_idx, (int, float, str)):
            bus_idx = [bus_idx]

        for item in bus_idx:
            idx = []
            for e, b in enumerate(self.bus):
                if b == item:
                    idx.append(self.idx[e])
            if len(idx) == 1:
                idx = idx[0]
            elif len(idx) == 0:
                idx = None

            ret.append(idx)

        if len(ret) == 1:
            ret = ret[0]

        return ret

    def link_bus(self, bus_idx):
        """
        Return the indices of elements linking the given buses

        :param bus_idx:
        :return:
        """
        ret = []

        if not self._config['is_series']:
            self.log(
                'link_bus function is not valid for non-series model <{}>'.
                format(self.name))
            return []

        if isinstance(bus_idx, (int, float, str)):
            bus_idx = [bus_idx]

        fkey = list(self._ac.keys())
        if 'bus' in fkey:
            fkey.remove('bus')

        nfkey = len(fkey)
        fkey_val = [self.__dict__[i] for i in fkey]

        for item in bus_idx:
            idx = []
            key = []
            for i in range(self.n):
                for j in range(nfkey):
                    if fkey_val[j][i] == item:
                        idx.append(self.idx[i])
                        key.append(fkey[j])
                        # <= 1 terminal should connect to the same bus
                        break

            if len(idx) == 0:
                idx = None
            if len(key) == 0:
                key = None

            ret.append((idx, key))

        return ret

    def elem_find(self, field, value):
        """
        Return the indices of elements whose field first satisfies the given values

        ``value`` should be unique in self.field.
        This function does not check the uniqueness.

        :param field: name of the supplied field
        :param value: value of field of the elemtn to find
        :return: idx of the elements
        :rtype: list, int, float, str
        """
        if isinstance(value, (int, float, str)):
            value = [value]

        f = list(self.__dict__[field])
        uid = np.vectorize(f.index)(value)
        return self.get_idx(uid)

    def _check_Vn(self):
        """Check data consistency of Vn and Vdcn if connected to Bus or Node

        :return None
        """
        if hasattr(self, 'bus') and hasattr(self, 'Vn'):
            bus_Vn = self.read_data_ext('Bus', field='Vn', idx=self.bus)
            for name, bus, Vn, Vn0 in zip(self.name, self.bus, self.Vn,
                                          bus_Vn):
                if Vn != Vn0:
                    self.log(
                        '<{}> has Vn={} different from bus <{}> Vn={}.'.format(
                            name, Vn, bus, Vn0), WARNING)

        if hasattr(self, 'node') and hasattr(self, 'Vdcn'):
            node_Vdcn = self.read_data_ext('Node', field='Vdcn', idx=self.node)
            for name, node, Vdcn, Vdcn0 in zip(self.name, self.node, self.Vdcn,
                                               node_Vdcn):
                if Vdcn != Vdcn0:
                    self.log(
                        '<{}> has Vdcn={} different from node <{}> Vdcn={}.'
                        .format(name, Vdcn, node, Vdcn0), WARNING)

    def link_from(self):
        """

        Returns
        -------

        """
        pass

    def link_to(self, model, idx, self_idx):
        """
        Register (self.name, self.idx) in `model._from`

        Returns
        -------

        """
        if model in self.system.loaded_groups:

            # access group instance
            grp = self.system.__dict__[model]

            # doing it one by one
            for i, self_i in zip(idx, self_idx):
                # query model name and access model instance
                mdl_name = grp._idx_model[i]
                mdl = self.system.__dict__[mdl_name]

                # query the corresponding uid
                u = mdl.get_uid(i)

                # update `mdl_from`
                name_idx_pair = (self._name, self_i)
                if name_idx_pair not in mdl.mdl_from[u]:
                    mdl.mdl_from[u].append(name_idx_pair)

        else:
            # access model instance
            mdl = self.system.__dict__[model]
            uid = mdl.get_uid(idx)

            for u, self_i in zip(uid, self_idx):
                name_idx_pair = (self._name, self_i)

                if name_idx_pair not in mdl.mdl_from[u]:
                    mdl.mdl_from[u].append(name_idx_pair)

    def get_element_data(self, idx):
        """
        Get all data associated with an element whose index is `idx` in a dict.

        Returns
        -------
        dict : param, value pairs
        """
        out = {}
        if idx not in self.idx:
            logger.debug('Model {} does not have element {}'.format(self.name, idx))
            return out

        idx_int = self.get_uid(idx)

        out['idx'] = self.idx[idx_int]
        out['name'] = self.name[idx_int]

        for item in self._data.keys():
            out[item] = self.__dict__[item][idx_int]

        for item in self._algebs:
            out[item] = self.__dict__[item][idx_int]

        for item in self._states:
            out[item] = self.__dict__[item][idx_int]

        return out

    def build_service(self):
        """
        Prototype function for building service variables. This function is to be called during `init0`
        or `init1`. Must be overloaded by derived classes.

        Returns
        -------

        """
        pass

    def init1(self, dae):
        pass

    def gcall(self, dae):
        pass

    def fcall(self, dae):
        pass

    def fxcall(self, dae):
        pass

    def gycall(self, dae):
        pass

    # def var_store_snapshot(self, variable='all'):
    #     """
    #     Store a snapshot of variable values to self._snapshot.
    #
    #     :param variable: name of the variable, or ``all``
    #     :return: None
    #     """
    #
    #     # store (variable name, equation name) in the list ``var_eq_pairs``
    #     # equation name is in ('x', 'y')
    #
    #     var_eq_pairs = []
    #     dae = self.system.dae
    #     if variable != 'all':
    #         if variable in self._states:
    #             var_eq_pairs.append((variable, 'x'))
    #         elif variable in self._algebs:
    #             var_eq_pairs.append((variable, 'y'))
    #         else:
    #             raise NameError('<{}> is not a variable of model {}'
    #                             .format(variable, self._name))
    #     else:
    #         for var in self._states:
    #             var_eq_pairs.append((var, 'x'))
    #         for var in self._algebs:
    #             var_eq_pairs.append((var, 'y'))
    #
    #     # retrieve values and store in ``self._snapshot``
    #
    #     for var, eq in var_eq_pairs:
    #         idx = self.__dict__[var]
    #         self._snapshot[var] = dae.__dict__[eq][idx]



import os

from cvxopt import matrix

from andes.utils.tab import simpletab


def format_newline():
    return '\n'


def format_title(item):
    return item


def format_item(item, val):
    return '{:20s} {:s}'.format(item, str(val))


def format_table(header, data, title=None):
    # fmt = ".4g"
    # return tabulate(data, headers=header, floatfmt=fmt)
    table = simpletab(data=data, header=header)
    return table.draw()
    # return []
    # pass


def dump_data(text, header, rowname, data, file, width=14, precision=5):
    out = ''

    os.makedirs(os.path.abspath(os.path.dirname(file)), exist_ok=True)
    with open(file, 'w') as fid:

        for Text, Header, Rowname, Data in zip(text, header, rowname, data):
            # Write Text
            if Text:
                fid.writelines(
                    Text
                )

            # Write Header
            if Header:
                ncol = len(Header)
                s = ' ' * width
                s += '{:>{width}s}' * ncol + '\n'
                fid.writelines(s.format(*Header, width=width))  # Mind the asterisk
                fid.write('\n')

            # Append Rowname to Data
            # Data is a list of column lists
            if Rowname:
                ncol = 0
                for idx, item in enumerate(Rowname):  # write by row as always
                    if not Data:
                        out = ''
                    elif isinstance(Data[0], list):  # list of list in Data
                        ncol = len(Data)
                        out = [Data[i][idx] for i in range(ncol)]
                    elif isinstance(Data[0],
                                    (int, float)):  # Is just a list of numbers
                        ncol = 1
                        out = [Data[idx]]
                    elif isinstance(Data, (int, float)):
                        ncol = 1
                        out = [Data]
                    elif isinstance(Data, matrix):  # Data is a matrix
                        pass
                    else:
                        print('Unexpected Data during output, in formats/txt.py')

                    s = '{:{width}s}'  # for row header
                    for col in out:
                        if isinstance(col, (int, float)):
                            s += '{:{width}.{precision}g}'
                        elif isinstance(col, str):
                            if len(col) > width:
                                col = col[:width]
                            s += '{:{width}s}'
                        else:
                            pass
                    s += '\n'

                    fid.write(
                        s.format(
                            str(item), *out, width=width, precision=precision))
            fid.write('\n')



from .jit import JIT  # NOQA

__all__ = ['jits', 'non_jits', 'all_models', 'all_models_list']

order = [
    'Bus', 'Node', 'Ground', 'line', 'pq', 'pv', 'zone', 'shunt',
    'measurement', 'synchronous', 'governor', 'avr', 'pss', 'windturbine',
    'wind', 'coi', 'BArea', 'eAGC', 'AGC', 'AGCSyn', 'AGCTG',
    'R', 'L', 'C', 'RLs', 'RCs', 'RCp', 'RLCp', 'RLCs',
    'DCgen', 'vsc', 'AGCTGVSC', 'AGCSynVSC', 'AGCMPC', 'LoadRamp', 'Recorder',
]

non_jits = {
    'bus': {
        'Bus': 'Bus'
    },
    'pq': {
        'PQ': 'PQ'
    },
    'pv': {
        'PV': 'PV',
        'Slack': 'SW'
    },
    'line': {
        'Line': 'Line'
    },
    'shunt': {
        'Shunt': 'Shunt'
    },
    'zone': {
        'Zone': 'Zone',
        'Area': 'Area',
        'Region': 'Region',
    },
    'coi': {
        'COI': 'COI',
    },
    'dcbase': {
        'Node': 'Node',
        'Ground': 'Ground',
        'R': 'R',
        'L': 'L',
        'C': 'C',
        'RLs': 'RLs',
        'RCs': 'RCs',
        'RCp': 'RCp',
        'RLCp': 'RLCp',
        'RLCs': 'RLCs',
        'DCgen': 'DCgen',
        'DCload': 'DCload',
    },
    'synchronous': {
        'Syn2': 'Syn2',
        'Syn6a': 'Syn6a',
    },
    'governor': {
        'TG2': 'TG2',
        'TG1': 'TG1'
    },
    'measurement': {
        'BusFreq': 'BusFreq',
        'PMU': 'PMU',
    },
    'avr': {
        'AVR3': 'AVR3',
        'AVR2': 'AVR2',
        'AVR1': 'AVR1',
    },
    'pss': {
        'PSS1': 'PSS1',
        'PSS2': 'PSS2',
    },
    'fault': {
        'Fault': 'Fault'
    },
    'breaker': {
        'Breaker': 'Breaker'
    },
    'event': {
        'GenTrip': 'GenTrip',
        'LoadShed': 'LoadShed',
        'LoadScale': 'LoadScale',
        'LoadRamp': 'LoadRamp',
    },
    'recorder': {
        'Recorder': 'Recorder'
    },
}

jits = {
    'agc': {
        'BArea': 'BArea',
        'AGC': 'AGC',
        'AGCSyn': 'AGCSyn',
        'AGCTG': 'AGCTG',
        'AGCTGVSC': 'AGCTGVSC',
        'AGCSynVSC': 'AGCSynVSC',
        'AGCMPC': "AGCMPC",
        'eAGC': 'eAGC',
    },
    'vsc': {
        'VSC': 'VSC',
        'VSC1': 'VSC1',
        'VSC1_IE': 'VSC1_IE',
        'VSC1_IE2': 'VSC1_IE2',
        'VSC2A': 'VSC2A',
        'VSC2B': 'VSC2B'
    },
    'windturbine': {
        'WTG3': 'WTG3',
        'WTG4DC': 'WTG4DC',
    },
    'wind': {
        'Weibull': 'Weibull',
        'ConstWind': 'ConstWind'
    },
}

all_models = dict(jits)
all_models.update(non_jits)

all_models_list = []

for key in sorted(all_models.keys()):
    val = all_models[key]
    all_models_list.extend(sorted(list(val.values())))



"""
Timed event base class
"""

from cvxopt import matrix, mul, spmatrix
from andes.utils.math import zeros

from .base import ModelBase
import numpy.random


class EventBase(ModelBase):
    """
    Base class for timed events
    """

    def define(self):
        self.param_remove('Sn')
        self.param_remove('Vn')
        self._group = 'Event'
        self._last_time = -1

    def get_times(self):
        """
        Return a list of occurrance times of the events

        :return: list of times
        """
        if not self.n:
            return list()

        ret = list()

        for item in self._event_times:
            ret += list(self.__dict__[item])

        return ret + list(matrix(ret) - 1e-6)

    def is_time(self, t):
        """
        Return the truth value of whether ``t`` is a time in
        ``self.get_times()``

        :param float t: time to check for
        :return bool: truth value
        """
        if not self.n:
            return False
        else:
            return t in self.get_times()

    def apply(self, sim_time):
        """
        Apply the event and
        :param sim_time:
        :return:
        """
        if not self.n:
            return

        # skip if already applied
        if self._last_time != sim_time:
            self._last_time = sim_time
        else:
            return


class GenTrip(EventBase):
    """
    Timed generator trip and reconnect
    """

    def define(self):
        super(GenTrip, self).define()
        self.param_define(
            'gen',
            default=None,
            mandatory=True,
            descr='generator idx',
            tomatrix=False)
        self.param_define(
            't1',
            default=-1,
            mandatory=False,
            descr='generator trip time',
            event_time=True)
        self.param_define(
            't2',
            default=-1,
            mandatory=False,
            descr='generator reconnect time',
            event_time=True)

        self._init()

    def apply(self, sim_time):
        super(GenTrip, self).apply(sim_time)
        for i in range(self.n):
            gen_idx = self.gen[i]

            if self.t1[i] == sim_time:
                self.log('Applying generator trip on <{}> at t={}'.format(
                    gen_idx, sim_time))
                self.system.Synchronous.set_field('u', gen_idx, 0)
            if self.t2[i] == sim_time:
                self.log('Applying generator reconnect on <{}> at t={}'.format(
                    gen_idx, sim_time))
                self.system.Synchronous.set_field('u', gen_idx, 1)


class LoadShed(EventBase):
    """
    Timed load shedding and reconnect
    """

    def define(self):
        super(LoadShed, self).define()
        self.param_define(
            'load',
            default=None,
            mandatory=True,
            descr='load idx',
            tomatrix=False)
        self.param_define(
            'group',
            default='StaticLoad',
            mandatory=True,
            descr='load group, StaticLoad or DynLoad',
            tomatrix=False)
        self.param_define(
            't1',
            default=-1,
            mandatory=False,
            descr='load shedding time',
            event_time=True)
        self.param_define(
            't2',
            default=-1,
            mandatory=False,
            descr='load reconnect time',
            event_time=True)

        self._init()

    def apply(self, sim_time):
        super(LoadShed, self).apply(sim_time)

        for i in range(self.n):
            load_idx = self.load[i]
            group = self.group[i]

            if self.t1[i] == sim_time:
                self.log('Applying load shedding on <{}> at t={}'.format(
                    load_idx, sim_time))
                self.system.__dict__[group].set_field('u', load_idx, 0)
            if self.t2[i] == sim_time:
                self.log('Applying Gen reconnect on <{}> at t={}'.format(
                    load_idx, sim_time))
                self.system.__dict__[group].set_field('u', load_idx, 1)


class LoadScale(EventBase):
    """
    Timed load scaling or increment
    """

    def define(self):
        super(LoadScale, self).define()
        self.param_define(
            'load',
            default=None,
            mandatory=True,
            descr='load idx',
            tomatrix=False)
        self.param_define(
            'group',
            default='StaticLoad',
            mandatory=True,
            descr='load group, StaticLoad or DynLoad',
            tomatrix=False)
        self.param_define(
            't1',
            default=-1,
            mandatory=False,
            descr='load scaling time',
            event_time=True)
        self.param_define('scale', default=1, mandatory=False, descr='load scaling factor', to_matrix=True)
        self.param_define('inc', default=0, mandatory=False, descr='load increment value', to_matrix=True)
        self.param_define('rand', default=0, mandatory=False,
                          descr='Multiply the load inc by a Gaussian (0, 0.1) rand', to_matrix=True)

        self._rand_coeff = []
        self._init()

    def apply(self, sim_time):
        super(LoadScale, self).apply(sim_time)

        if len(self._rand_coeff) == 0:
            self._rand_coeff = numpy.random.rand(self.n)

        for i in range(self.n):
            load_idx = self.load[i]
            group = self.group[i]

            if self.t1[i] == sim_time:
                self.log('Applying load scaling on <{}> at t={}'.format(
                    load_idx, sim_time))
                old_load = self.system.__dict__[group].get_field('p', load_idx)

                new_load = mul(old_load, self.scale[i])
                if self.rand[i]:
                    new_load += self.inc[i] * self._rand_coeff[i]
                else:
                    new_load += self.inc[i]

                self.system.__dict__[group].set_field('p', load_idx, new_load)


class LoadRamp(ModelBase):
    """
    Continuous load ramping
    """
    def define(self):
        self.param_remove('Sn')
        self.param_remove('Vn')
        self._group = 'Event'

        self.param_define('load', default=None, mandatory=True, descr="load idx", tomatrix=False)
        self.param_define('group', default='StaticLoad', mandatory=True,
                          descr="load group, StaticLoad or DynLoad", tomatrix=False)
        self.param_define('t1', default=-1, mandatory=True, descr='start time', tomatrix=True)
        self.param_define('t2', default=-1, mandatory=True, descr='end time', tomatrix=True)
        self.param_define('p_rate', default=1, mandatory=False, descr='rate of ramping per hour in percentage',
                          tomatrix=True)
        self.param_define('p_amount', default=0, mandatory=False, descr='the amount of ramping per hour in pu',
                          tomatrix=True)
        self.param_define('q_rate', default=1, mandatory=False, descr='rate of ramping per hour in percentage',
                          tomatrix=True)
        self.param_define('q_amount', default=0, mandatory=False, descr='the amount of ramping per hour in pu',
                          tomatrix=True)

        self.service_define("p0", matrix)
        self.service_define("q0", matrix)

        self.service_define("p_out", matrix)
        self.service_define("q_out", matrix)

        self.calls.update({'gcall': True,
                           'init1': True,
                           })

        self._init()

    def init1(self, dae):
        # check the exclusivity of rate and amount

        # obtain the p0 and q0 at the time of the start
        self.copy_data_ext("StaticLoad", field="p0", dest="p0", idx=self.load)
        self.copy_data_ext("StaticLoad", field="q0", dest="q0", idx=self.load)
        self.copy_data_ext("StaticLoad", field="a", dest="a", idx=self.load)
        self.copy_data_ext("StaticLoad", field="v", dest="v", idx=self.load)

        self.p_out = zeros(self.n, 1)
        self.q_out = zeros(self.n, 1)

    def gcall(self, dae):
        # call the function to calculate the load (p and q) at the present time
        if dae.t < 0:
            return
        # calculate the load increase
        self.calc_p(dae.t)
        self.calc_q(dae.t)

        # apply the load change to the bus equations
        dae.g += spmatrix(self.p_out, self.a, [0] * self.n, (dae.m, 1), 'd')
        dae.g += spmatrix(self.q_out, self.v, [0] * self.n, (dae.m, 1), 'd')

    def calc_p(self, t):
        for i in range(self.n):
            if t < self.t1[i]:
                self.p_out[i] = 0
            elif t > self.t2[i]:
                continue
            else:
                self.p_out[i] = (self.p_rate[i] * self.p0[i] / 60 / 60) * (t - self.t1[i]) +\
                                (self.p_amount[i] / 60 / 60) * (t - self.t1[i])

    def calc_q(self, t):
        for i in range(self.n):
            if t < self.t1[i]:
                self.q_out[i] = 0
            elif t > self.t2[i]:
                continue
            else:
                self.q_out[i] = (self.q_rate[i] * self.q0[i] / 60 / 60) * (t - self.t1[i]) +\
                                (self.q_amount[i] / 60 / 60) * (t - self.t1[i])

    def get_times(self):
        return []

    def is_time(self, t):
        return False



import numpy as np
from cvxopt import matrix  # NOQA
from cvxopt import mul, div  # NOQA
from scipy.sparse import coo_matrix

from .base import ModelBase
from ..consts import Gy  # NOQA
from ..utils.math import altb, agtb, aorb, nota
from ..utils.math import zeros, ones


class PQ(ModelBase):
    """Static PQ load class"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'PQ'
        self._group = 'StaticLoad'
        self._category = 'Load'
        self._data.update({
            'bus': None,
            'p': 0,
            'q': 0,
            'owner': 0,
            'vmax': 1.1,
            'vmin': 0.9,
        })
        self._units.update({
            'bus': 'na',
            'p': 'pu',
            'owner': 'na',
            'vmax': 'pu',
            'vmin': 'pu',
        })
        self._params.extend(['p', 'q', 'vmax', 'vmin'])
        self._descr.update({
            'bus': 'bus number',
            'p': 'constant p value',
            'q': 'constant q value',
            'owner': 'owner code',
            'vmax': 'max voltage before switching to Z',
            'vmin': 'min voltage before switching to Z'
        })
        self._ac = {'bus': ['a', 'v']}
        self._powers = ['p', 'q']
        self._service = ['p0', 'q0', 'v0', 'below', 'above',
                         'normal']  # p0 and q0 are used during computation
        self.calls.update({
            'gcall': True,
            'gycall': True,
            'init0': True,
            'init1': True,
            'pflow': True,
            'shunt': True,
        })
        self._init()

    def init0(self, dae):
        """Set initial p and q for power flow"""
        self.p0 = matrix(self.p, (self.n, 1), 'd')
        self.q0 = matrix(self.q, (self.n, 1), 'd')

    def init1(self, dae):
        """Set initial voltage for time domain simulation"""
        self.v0 = matrix(dae.y[self.v])

    def gcall(self, dae):
        k = ones(self.n, 1)

        if self.system.config.forcez:
            if self.v0:
                k = div(dae.y[self.v]**2, self.v0**2)
        elif self.system.config.forcepq:
            pass
        else:
            k = zeros(self.n, 1)

            self.below = altb(dae.y[self.v], self.vmin)
            k += mul(self.below, div(dae.y[self.v]**2, self.vmin**2))

            self.above = agtb(dae.y[self.v], self.vmax)
            k += mul(self.above, div(dae.y[self.v]**2, self.vmax**2))

            normal = nota(aorb(self.below, self.above))
            k += mul(normal, ones(self.n, 1))

        k = mul(self.u, k)
        self.p0 = mul(k, self.p)
        self.q0 = mul(k, self.q)

        if self.n <= 400:
            for a, v, p, q in zip(self.a, self.v, self.p0, self.q0):
                dae.g[a] += p
                dae.g[v] += q
        else:
            # use scipy.sparse.coo_matrix to accelerate for large systems

            p0 = coo_matrix(
                (np.array(self.p0).reshape((-1)), (self.a, np.zeros(self.n))),
                shape=(dae.m, 1))
            q0 = coo_matrix(
                (np.array(self.q0).reshape((-1)), (self.v, np.zeros(self.n))),
                shape=(dae.m, 1))

            dae.g += matrix(p0.toarray())
            dae.g += matrix(q0.toarray())

    def gycall(self, dae):
        k = zeros(self.n, 1)
        if self.system.config.forcepq:
            return
        elif self.system.config.forcez:
            if self.v0:
                k = div(2 * dae.y[self.v], self.v0**2)
        else:
            k += mul(self.below, div(2 * dae.y[self.v], self.vmin**2))
            k += mul(self.above, div(2 * dae.y[self.v], self.vmax**2))
        k = mul(self.u, k)

        dae.add_jac(Gy, mul(self.p, k), self.a, self.v)
        dae.add_jac(Gy, mul(self.q, k), self.v, self.v)



import logging

from . import ConfigBase
from ..utils.cached import cached

logger = logging.getLogger(__name__)


class Pflow(ConfigBase):
    def __init__(self, **kwargs):
        self.flatstart = False
        self.tol = 1e-6
        self.maxit = 100
        self.pv2pq = False
        self.ipv2pq = 4
        self.npv2pq = 1
        self.report = 'default'
        self.show = True
        self.method = 'NR'
        self.method_alt = ['NR', 'FDPF', 'FDBX', 'FDXB', 'DCPF']
        self.sortbuses = 'data'
        self.sortbuses_alt = ['data', 'idx']
        self.static = False
        self.switch2nr = False
        self.units = 'pu'
        self.units_alt = ['pu', 'nominal']
        self.usedegree = True
        super(Pflow, self).__init__(**kwargs)

    @cached
    def config_descr(self):
        descriptions = {
            'flatstart': 'flat start for power flow problem',
            'maxit': 'the maximum iteration number',
            'pv2pq': 'check Q limit and convert PV to PQ',
            'ipv2pq': 'the interation from which to convert PV to PQ',
            'npv2pq': 'the maximum number of PVs to convert in one iteration',
            'method': 'power flow routine solver type',
            'sortbuses': 'index to sort buses',
            'switch2nr': 'switch to Newton Raphson method if non-convergence',
            'units': 'the unit for the power flow report',
            'usedegree': 'use degree in the power flow report',
        }
        return descriptions

    def check(self):
        if self.method not in self.method_alt:
            self.method = 'NR'
        if self.method == 'DCPF':
            self.flatstart = True
        if self.pv2pq is True:
            logger.info('PV Qlim check when iter >= {} or mismatch <= {}'.
                        format(int(self.ipv2pq), 1e4*self.tol))
        return True



from decimal import Decimal, ROUND_DOWN
from time import time


def elapsed(t0=0.0):
    """get elapsed time from the give time

    Returns:
        now: the absolute time now
        dt_str: elapsed time in string
        """
    now = time()
    dt = now - t0
    dt_sec = Decimal(str(dt)).quantize(Decimal('.0001'), rounding=ROUND_DOWN)
    if dt_sec <= 1:
        dt_str = str(dt_sec) + ' second'
    else:
        dt_str = str(dt_sec) + ' seconds'
    return now, dt_str



import numpy as np
from cvxopt import matrix, mul, spmatrix
from scipy.sparse import coo_matrix

from .base import ModelBase
from ..consts import Gy  # NOQA


class Shunt(ModelBase):
    """Static shunt device"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'StaticVar'
        self._name = 'Shunt'
        self._data.update({
            'bus': None,
            'g': 0,
            'b': 0,
            'fn': 60.0,
        })
        self._units.update({
            'bus': 'na',
            'g': 'pu',
            'b': 'pu',
            'fn': 'Hz',
        })
        self._params.extend([
            'g',
            'b',
            'fn',
        ])
        self._descr.update({
            'bus': 'idx of connected bus',
            'g': 'shunt conductance (real part)',
            'b': 'shunt susceptance (positive as capatance)',
            'fn': 'rated frequency',
        })
        self._ac = {'bus': ['a', 'v']}
        self.calls.update({
            'gcall': True,
            'gycall': True,
            'pflow': True,
            'shunt': True,
        })

        self._y = ['g', 'b']
        self._init()

    def full_y(self, Y):
        """Add self(shunt) into full Jacobian Y"""
        if not self.n:
            return
        Ysh = matrix(self.g,
                     (self.n, 1), 'd') + 1j * matrix(self.b, (self.n, 1), 'd')
        uYsh = mul(self.u, Ysh)
        Y += spmatrix(uYsh, self.a, self.a, Y.size, 'z')

    def gcall(self, dae):
        # scipy.sparse.coo_matrix speed up for large matrices ======

        v2 = mul(self.u, dae.y[self.v]**2)
        p_inj = mul(v2, self.g)
        q_inj = -mul(v2, self.b)

        if self.n <= 400:
            for a, v, p, q in zip(self.a, self.v, p_inj, q_inj):
                dae.g[a] += p
                dae.g[v] += q
        else:
            p = np.array(p_inj).reshape((-1))
            q = np.array(q_inj).reshape((-1))

            p = coo_matrix(
                (p, (self.a, np.zeros(self.n))), shape=(dae.m, 1)).toarray()
            q = coo_matrix(
                (q, (self.v, np.zeros(self.n))), shape=(dae.m, 1)).toarray()

            dae.g += matrix(p)
            dae.g += matrix(q)

    def gycall(self, dae):
        dV2 = mul(self.u, 2 * dae.y[self.v])
        dPdV = mul(dV2, matrix(self.g, (self.n, 1), 'd'))
        dQdV = -mul(dV2, matrix(self.b, (self.n, 1), 'd'))

        dae.add_jac(Gy, dPdV, self.a, self.v)
        dae.add_jac(Gy, dQdV, self.v, self.v)



from scipy.sparse import csc_matrix
from scipy.sparse.linalg import spsolve


from cvxopt import umfpack, matrix

import numpy as np

try:
    from cvxoptklu import klu
    KLU = True
except ImportError:
    KLU = False


class Solver(object):
    """
    Sparse matrix solver class. Wraps UMFPACK and KLU with a unified interface.
    """
    def __init__(self, sparselib='umfpack'):
        self.sparselib = sparselib

    def symbolic(self, A):
        """
        Return the symbolic factorization of sparse matrix ``A``

        Parameters
        ----------
        sparselib
            Library name in ``umfpack`` and ``klu``
        A
            Sparse matrix

        Returns
        symbolic factorization
        -------

        """

        if self.sparselib == 'umfpack':
            return umfpack.symbolic(A)

        elif self.sparselib == 'klu':
            return klu.symbolic(A)

        elif self.sparselib in ('spsolve', 'cupy'):
            raise NotImplementedError

    def numeric(self, A, F):
        """
        Return the numeric factorization of sparse matrix ``A`` using symbolic factorization ``F``

        Parameters
        ----------
        A
            Sparse matrix
        F
            Symbolic factorization

        Returns
        -------
        N
            Numeric factorization of ``A``
        """
        if self.sparselib == 'umfpack':
            return umfpack.numeric(A, F)

        elif self.sparselib == 'klu':
            return klu.numeric(A, F)

        elif self.sparselib in ('spsolve', 'cupy'):
            raise NotImplementedError

    def solve(self, A, F, N, b):
        """
        Solve linear system ``Ax = b`` using numeric factorization ``N`` and symbolic factorization ``F``.
        Store the solution in ``b``.

        Parameters
        ----------
        A
            Sparse matrix
        F
            Symbolic factorization
        N
            Numeric factorization
        b
            RHS of the equation

        Returns
        -------
        None
        """
        if self.sparselib == 'umfpack':
            umfpack.solve(A, N, b)
            return b

        elif self.sparselib == 'klu':
            klu.solve(A, F, N, b)
            return b

        elif self.sparselib in ('spsolve', 'cupy'):
            raise NotImplementedError

    def linsolve(self, A, b):
        """
        Solve linear equation set ``Ax = b`` and store the solutions in ``b``.

        Parameters
        ----------
        A
            Sparse matrix

        b
            RHS of the equation

        Returns
        -------
        None
        """
        if self.sparselib == 'umfpack':
            umfpack.linsolve(A, b)
            return b

        elif self.sparselib == 'klu':
            klu.linsolve(A, b)
            return b

        elif self.sparselib in ('spsolve', 'cupy'):
            ccs = A.CCS
            size = A.size
            data = np.array(ccs[2]).reshape((-1,))
            indices = np.array(ccs[1]).reshape((-1,))
            indptr = np.array(ccs[0]).reshape((-1,))

            A = csc_matrix((data, indices, indptr), shape=size)

            if self.sparselib == 'spsolve':
                x = spsolve(A, b)
                return matrix(x)

            elif self.sparselib == 'cupy':
                # delayed import for startup speed
                import cupy as cp  # NOQA
                from cupyx.scipy.sparse import csc_matrix as csc_cu  # NOQA
                from cupyx.scipy.sparse.linalg.solve import lsqr as cu_lsqr  # NOQA

                cu_A = csc_cu(A)
                cu_b = cp.array(np.array(b).reshape((-1,)))
                x = cu_lsqr(cu_A, cu_b)

                return matrix(cp.asnumpy(x[0]))



"""Synchronous generator classes"""

from cvxopt import matrix, spmatrix  # NOQA
from cvxopt import mul, div, log, sin, cos  # NOQA

from .base import ModelBase
from ..consts import Fx, Fy, Gx, Gy  # NOQA
from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..consts import jpi2
from ..utils.math import polar, conj, exp


class SynBase(ModelBase):
    """Base class for synchronous generators"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._group = 'Synchronous'
        self._data.update({
            'fn': 60.0,
            'bus': None,
            'D': 0.0,
            'M': 6,
            'ra': 0.0,
            'xl': 0.0,
            'xq': 1.7,
            'gammap': 1.0,
            'gammaq': 1.0,
            'coi': None,
            'gen': None,
            'kp': 0,
            'kw': 0,
            'S10': 0,
            'S12': 0,
        })
        self._params.extend([
            'D',
            'M',
            'ra',
            'xl',
            'xq',
            'gammap',
            'gammaq',
            'kp',
            'kw',
            'S10',
            'S12',
        ])
        self._descr.update({
            'fn':
            'rated frequency',
            'bus':
            'interface bus id',
            'D':
            'Damping coefficient',
            'M':
            'machine start up time (2H)',
            'ra':
            'armature resistance',
            'xl':
            'leakage reactance',
            'xq':
            'q-axis synchronous reactance',
            'gammap':
            'active power ratio of all generators on this bus',
            'gammaq':
            'reactive power ratio',
            'coi':
            'center of inertia index',
            'gen':
            'static generator index',
            'kp':
            'active power feedback gain',
            'kw':
            'speed feedback gain',
            'S10':
            'first saturation factor',
            'S12':
            'second saturation factor',
        })
        self._units.update({
            'M': 'MWs/MVA',
            'D': 'pu',
            'fn': 'Hz',
            'ra': 'ohm',
            'xd': 'ohm',
            'gammap': 'pu',
            'gammaq': 'pu',
        })
        self.calls.update({
            'init1': True,
            'dyngen': True,
            'gcall': True,
            'gycall': True,
            'fcall': True,
            'fxcall': True,
            'jac0': True,
        })
        self._ac = {'bus': ['a', 'v']}
        self._states = ['delta', 'omega']
        self._fnamex = ['\\delta', '\\omega']
        self._algebs = ['p', 'q', 'pm', 'vf', 'Id', 'Iq', 'vd', 'vq']
        self._fnamey = ['P', 'Q', 'P_m', 'V_f', 'I_d', 'I_q', 'V_d', 'V_q']
        self._powers = ['M', 'D']
        self._z = ['ra', 'xl', 'xq']
        self._zeros = ['M']
        self._mandatory = ['bus', 'gen']
        self._service = ['pm0', 'vf0', 'c1', 'c2', 'c3', 'ss', 'cc', 'iM']

    def build_service(self):
        """Build service variables"""
        self.iM = div(1, self.M)

    def set_vf0(self, vf):
        """set value for self.vf0 and dae.y[self.vf]"""
        self.vf0 = vf
        self.system.dae.y[self.vf] = matrix(vf)

    def init1(self, dae):
        self.system.rmgen(self.gen)
        self.build_service()

        p0 = mul(self.u, self.system.Bus.Pg[self.a], self.gammap)
        q0 = mul(self.u, self.system.Bus.Qg[self.a], self.gammaq)
        v0 = mul(1, dae.y[self.v])
        theta0 = dae.y[self.a]
        v = polar(v0, theta0)
        S = p0 - q0 * 1j
        I = div(S, conj(v))  # NOQA
        E = v + mul(self.ra + self.xq * 1j, I)

        dae.y[self.p] = p0
        dae.y[self.q] = q0

        delta = log(div(E, abs(E) + 0j))
        dae.x[self.delta] = mul(self.u, delta.imag())
        dae.x[self.omega] = matrix(self.u, (self.n, 1), 'd')

        # d- and q-axis voltages and currents
        vdq = mul(self.u, mul(v, exp(jpi2 - delta)))
        idq = mul(self.u, mul(I, exp(jpi2 - delta)))
        vd = dae.y[self.vd] = vdq.real()
        vq = dae.y[self.vq] = vdq.imag()
        Id = dae.y[self.Id] = idq.real()
        Iq = dae.y[self.Iq] = idq.imag()

        # electro-mechanical torques / powers
        self.pm0 = mul(vq + mul(self.ra, Iq), Iq) + mul(
            vd + mul(self.ra, Id), Id)
        dae.y[self.pm] = self.pm0

    def gcall(self, dae):
        nzeros = [0] * self.n
        v = mul(self.u, dae.y[self.v])
        vd = mul(self.u, dae.y[self.vd])
        vq = mul(self.u, dae.y[self.vq])
        Id = mul(self.u, dae.y[self.Id])
        Iq = mul(self.u, dae.y[self.Iq])
        self.ss = sin(dae.x[self.delta] - dae.y[self.a])
        self.cc = cos(dae.x[self.delta] - dae.y[self.a])

        dae.g -= spmatrix(dae.y[self.p], self.a, nzeros, (dae.m, 1), 'd')
        dae.g -= spmatrix(dae.y[self.q], self.v, nzeros, (dae.m, 1), 'd')
        dae.g -= spmatrix(vd - mul(v, self.ss), self.vd, nzeros, (dae.m, 1),
                          'd')  # note d(vd)/d(delta)
        dae.g -= spmatrix(vq - mul(v, self.cc), self.vq, nzeros, (dae.m, 1),
                          'd')  # note d(vq)/d(delta)
        dae.g += spmatrix(
            mul(self.u, vd, Id) + mul(self.u, vq, Iq) - dae.y[self.p], self.p,
            nzeros, (dae.m, 1), 'd')
        dae.g += spmatrix(
            mul(self.u, vq, Id) - mul(self.u, vd, Iq) - dae.y[self.q], self.q,
            nzeros, (dae.m, 1), 'd')
        dae.g += spmatrix(dae.y[self.pm] - mul(self.u, self.pm0), self.pm,
                          nzeros, (dae.m, 1), 'd')
        dae.g += spmatrix(dae.y[self.vf] - mul(self.u, self.vf0), self.vf,
                          nzeros, (dae.m, 1), 'd')

    def saturation(self, e1q):
        """Saturation characteristic function"""
        return e1q

    def fcall(self, dae):
        dae.f[self.delta] = mul(self.u, self.system.wb,
                                dae.x[self.omega] - 1)

    def jac0(self, dae):
        dae.add_jac(Gy0, -self.u, self.a, self.p)
        dae.add_jac(Gy0, -self.u, self.v, self.q)
        dae.add_jac(Gy0, -self.u + 1e-6, self.vd, self.vd)
        dae.add_jac(Gy0, -self.u + 1e-6, self.vq, self.vq)
        dae.add_jac(Gy0, -1.0, self.p, self.p)
        dae.add_jac(Gy0, -1.0, self.q, self.q)
        dae.add_jac(Gy0, 1.0, self.pm, self.pm)
        dae.add_jac(Gy0, 1.0, self.vf, self.vf)

        dae.add_jac(Fx0, 1e-6, self.delta, self.delta)
        dae.add_jac(Fx0, mul(self.u, self.system.wb), self.delta,
                    self.omega)

    def gycall(self, dae):
        dae.add_jac(Gy, mul(self.u, dae.y[self.Id]), self.p, self.vd)
        dae.add_jac(Gy, mul(self.u, dae.y[self.Iq]), self.p, self.vq)
        dae.add_jac(Gy, mul(self.u, dae.y[self.vd]), self.p, self.Id)
        dae.add_jac(Gy, mul(self.u, dae.y[self.vq]), self.p, self.Iq)

        dae.add_jac(Gy, -mul(self.u, dae.y[self.Iq]), self.q, self.vd)
        dae.add_jac(Gy, mul(self.u, dae.y[self.Id]), self.q, self.vq)
        dae.add_jac(Gy, mul(self.u, dae.y[self.vq]), self.q, self.Id)
        dae.add_jac(Gy, -mul(self.u, dae.y[self.vd]), self.q, self.Iq)

        dae.add_jac(Gy, -mul(self.u, dae.y[self.v], self.cc), self.vd, self.a)
        dae.add_jac(Gy, mul(self.u, self.ss), self.vd, self.v)

        dae.add_jac(Gy, mul(self.u, dae.y[self.v], self.ss), self.vq, self.a)
        dae.add_jac(Gy, mul(self.u, self.cc), self.vq, self.v)

    def fxcall(self, dae):
        dae.add_jac(Gx, mul(self.u, dae.y[self.v], self.cc), self.vd,
                    self.delta)
        dae.add_jac(Gx, -mul(self.u, dae.y[self.v], self.ss), self.vq,
                    self.delta)


class Ord2(SynBase):
    """2nd order classical model"""

    def __init__(self, system, name):
        super().__init__(system, name)
        self._name = 'Syn2'
        self._data.update({'xd1': 1.9})
        self._params.extend(['xd1'])
        self._descr.update({'xd1': 'synchronous reactance'})
        self._units.update({'xd1': 'ohm'})
        self._z.extend(['xd1'])

    def build_service(self):
        super(Ord2, self).build_service()
        self.xq = self.xd1

    def init1(self, dae):
        super().init1(dae)
        self.set_vf0(dae.y[self.vq] + mul(self.ra, dae.y[self.Iq]) +
                     mul(self.xd1, dae.y[self.Id]))

    def gcall(self, dae):
        super(Ord2, self).gcall(dae)
        nzeros = [0] * self.n
        dae.g += spmatrix(
            mul(self.xd1, dae.y[self.Id]) - dae.y[self.vf], self.Id, nzeros,
            (dae.m, 1), 'd')
        dae.g += spmatrix(
            mul(self.xd1, dae.y[self.Iq]), self.Iq, nzeros, (dae.m, 1), 'd')

    def jac0(self, dae):
        super(Ord2, self).jac0(dae)
        dae.add_jac(Gy0, self.xd1, self.Id, self.Id)
        dae.add_jac(Gy0, -1.0, self.Id, self.vf)

        dae.add_jac(Gy0, self.xd1, self.Iq, self.Iq)


class Ord6a(SynBase):
    """6th order the Marconato's Model"""

    # todo: check e1d
    def __init__(self, system, name):
        super(Ord6a, self).__init__(system, name)
        self._name = 'Syn6a'
        self._data.update({
            'xd': 1.9,
            'xd1': 0.302,
            'xq1': 0.5,
            'xd2': 0.204,
            'xq2': 0.3,
            'Td10': 8.0,
            'Tq10': 0.8,
            'Td20': 0.04,
            'Tq20': 0.02,
            'Taa': 0.0,
            'S10': 0,
            'S12': 0
        })
        self._params.extend([
            'xd', 'xd1', 'xq1', 'xd2', 'xq2', 'Td10', 'Tq10', 'Td20', 'Tq20',
            'Taa', 'S10', 'S12'
        ])
        self._descr.update({
            'xd': 'd-axis synchronous reactance',
            'xd1': 'd-axis transient reactance',
            'xq1': 'q-axis transient reactance',
            'xd2': 'd-axis sub-transient reactance',
            'xq2': 'q-axis sub-transient reactance',
            'Td10': 'd-axis transient time constant',
            'Tq10': 'q-axis transient time constant',
            'Td20': 'd-axis sub-transient time constant',
            'Tq20': 'q-axis sub-transient time constant',
            'Taa': 'd-axis additional leakage time constant',
        })
        self._units.update({
            'xd': 'ohm',
            'xd1': 'ohm',
            'xd2': 'ohm',
            'xq1': 'ohm',
            'xq2': 'ohm',
            'Td10': 's',
            'Tq10': 's',
            'Td20': 's',
            'Tq20': 's',
            'Taa': 's',
            'S10': 'n/a',
            'S12': 'n/a'
        })
        self._mandatory.extend([
            'xd', 'xq', 'xd1', 'xq1', 'xd2', 'xq2', 'Td10', 'Tq10', 'Td20',
            'Tq20'
        ])
        self._service.extend([
            'c1', 'c2', 'c3', 'a1', 'a2', 'a3', 'a4', 'a5', 'a6', 'b1', 'b2',
            'b3', 'b4'
        ])
        self._states.extend(['e1d', 'e1q', 'e2d', 'e2q'])
        self._fnamex.extend(['e\'_d', 'e\'_q', 'e\'\'_d', 'e\'\'_q'])
        self._z.extend(['xd', 'xd1', 'xq1', 'xd2', 'xq2'])
        self._times.extend(['Td10', 'Tq10', 'Td20', 'Tq20', 'Taa'])

    def build_service(self):
        super(Ord6a, self).build_service()
        K = div(1, self.ra**2 + self.xd2**2)
        self.gd = mul(
            div(self.xd2, self.xd1), div(self.Td20, self.Td10),
            self.xd - self.xd1)
        self.gq = mul(
            div(self.xq2, self.xq1), div(self.Tq20, self.Tq10),
            self.xq - self.xq1)
        self.c1 = mul(self.ra, K)
        self.c2 = mul(self.xd2, K)
        self.c3 = mul(self.xq2, K)
        self.a1 = div(self.u, self.Td20)
        self.a2 = mul(self.a1, self.xd1 - self.xd2 + self.gd)
        self.a3 = mul(self.u, div(self.Taa, mul(self.Td10, self.Td20)))
        self.a4 = div(self.u, self.Td10)
        self.a5 = mul(self.a4, self.xd - self.xd1 - self.gd)
        self.a6 = mul(self.a4, 1 - div(self.Taa, self.Td10))
        self.b1 = div(self.u, self.Tq20)
        self.b2 = mul(self.b1, self.xq1 - self.xq2 + self.gq)
        self.b3 = div(self.u, self.Tq10)
        self.b4 = mul(self.b3, self.xq - self.xq1 - self.gq)

    def init1(self, dae):
        super(Ord6a, self).init1(dae)
        vd = mul(self.u, dae.y[self.vd])
        vq = mul(self.u, dae.y[self.vq])
        Id = mul(self.u, dae.y[self.Id])
        Iq = mul(self.u, dae.y[self.Iq])

        k1 = self.xd - self.xd1 - self.gd
        k2 = self.xd1 - self.xd2 + self.gd
        kq = self.xq - self.xq1 - self.gq
        kt = div(self.Taa, self.Td10)

        dae.x[self.e2q] = vq + mul(self.ra, Iq) + mul(self.xd2, Id)
        dae.x[self.e2d] = vd + mul(self.ra, Id) - mul(self.xq2, Iq)
        dae.x[self.e1d] = mul(Iq, kq)
        dae.x[self.e1q] = dae.x[self.e2q] + mul(k2, Id) - mul(
            kt,
            mul(k1 + k2, Id) + dae.x[self.e2q])
        self.set_vf0(
            div(mul(k1, Id) + self.saturation(dae.x[self.e1q]), 1 - kt))

    def gcall(self, dae):
        super(Ord6a, self).gcall(dae)
        dae.g[self.Id] += mul(self.u, self.xd2, dae.y[self.Id]) - mul(
            self.u, dae.x[self.e2q])
        dae.g[self.Iq] += mul(self.u, self.xq2, dae.y[self.Iq]) + mul(
            self.u, dae.x[self.e2d])

    def fcall(self, dae):
        super(Ord6a, self).fcall(dae)
        Id = mul(self.u, dae.y[self.Id])
        Iq = mul(self.u, dae.y[self.Iq])
        dae.f[self.e1d] = -mul(self.b3, dae.x[self.e1d], self.u) + mul(
            self.b4, Iq, self.u)
        dae.f[self.e1q] = -mul(
            self.saturation(dae.x[self.e1q]), self.a4, self.u) - mul(
                self.a5, Id, self.u) + mul(self.a6, dae.y[self.vf], self.u)
        dae.f[self.e2d] = -mul(self.b1, dae.x[self.e2d], self.u) + mul(
            self.b1, dae.x[self.e1d], self.u) + mul(self.b2, Iq, self.u)
        dae.f[self.e2q] = -mul(self.a1, dae.x[self.e2q], self.u) + mul(
            self.a1, dae.x[self.e1q], self.u) - mul(self.a2, Id, self.u) + mul(
                self.a3, dae.y[self.vf], self.u)

    def jac0(self, dae):
        super(Ord6a, self).jac0(dae)
        dae.add_jac(Gy0, mul(self.u, self.xd2) + 1e-6, self.Id, self.Id)
        dae.add_jac(Gx0, -self.u, self.Id, self.e2q)

        dae.add_jac(Gy0, mul(self.u, self.xq2) + 1e-6, self.Iq, self.Iq)
        dae.add_jac(Gx0, self.u, self.Iq, self.e2d)

        dae.add_jac(Fx0, -mul(self.u, self.b3) + 1e-6, self.e1d, self.e1d)
        dae.add_jac(Fy0, mul(self.u, self.b4), self.e1d, self.Iq)

        dae.add_jac(Fx0, -mul(self.u, self.a4) + 1e-6, self.e1q, self.e1q)
        dae.add_jac(Fy0, mul(self.u, self.a6), self.e1q, self.vf)
        dae.add_jac(Fy0, -mul(self.u, self.a5), self.e1q, self.Id)

        dae.add_jac(Fx0, -mul(self.u, self.b1) + 1e-6, self.e2d, self.e2d)
        dae.add_jac(Fx0, mul(self.u, self.b1), self.e2d, self.e1d)
        dae.add_jac(Fy0, mul(self.u, self.b2), self.e2d, self.Iq)

        dae.add_jac(Fx0, -mul(self.u, self.a1) + 1e-6, self.e2q, self.e2q)
        dae.add_jac(Fx0, mul(self.u, self.a1), self.e2q, self.e1q)
        dae.add_jac(Fy0, mul(self.u, self.a3), self.e2q, self.vf)
        dae.add_jac(Fy0, -mul(self.u, self.a2), self.e2q, self.Id)


class Flux0(object):
    """The simplified flux model as an appendix to generator models.
         0 = ra*id + psiq + vd
         0 = ra*iq - psid + vq
    """

    def __init__(self):
        self._algebs.extend(['psid', 'psiq'])
        self._fnamey.extend(['\\psi_d', '\\psi_q'])
        self._init()

    def init1(self, dae):
        dae.y[
            self.psiq] = -mul(self.u, self.ra, dae.y[self.Id]) - dae.y[self.vd]
        dae.y[self.psid] = mul(self.u, self.ra,
                               dae.y[self.Iq]) + dae.y[self.vq]

    def gcall(self, dae):
        dae.g[self.psiq] = mul(self.u, self.ra,
                               dae.y[self.Id]) + dae.y[self.psiq] + mul(
                                   self.u, dae.y[self.vd])
        dae.g[self.psid] = mul(self.u, self.ra,
                               dae.y[self.Iq]) - dae.y[self.psid] + mul(
                                   self.u, dae.y[self.vq])
        dae.g[self.Id] += mul(self.u, dae.y[self.psid])
        dae.g[self.Iq] += mul(self.u, dae.y[self.psiq])

    def gycall(self, dae):
        dae.add_jac(Gy, mul(self.u, self.ra), self.psiq, self.Id)
        dae.add_jac(Gy, mul(self.u, self.ra), self.psid, self.Iq)

    def fcall(self, dae):
        dae.f[self.omega] = mul(
            self.u, self.iM,
            dae.y[self.pm] - mul(dae.y[self.psid], dae.y[self.Iq]) + mul(
                dae.y[self.psiq], dae.y[self.Id]) - mul(
                    self.D, dae.x[self.omega] - 1))

    def fxcall(self, dae):
        dae.add_jac(Fy, mul(self.u, dae.y[self.Id], self.iM), self.omega,
                    self.psiq)
        dae.add_jac(Fy, mul(self.u, dae.y[self.psiq], self.iM), self.omega,
                    self.Id)
        dae.add_jac(Fy, -mul(self.u, dae.y[self.Iq], self.iM), self.omega,
                    self.psid)
        dae.add_jac(Fy, -mul(self.u, dae.y[self.psid], self.iM), self.omega,
                    self.Iq)

    def jac0(self, dae):
        dae.add_jac(Gy0, 1.0, self.psiq, self.psiq)
        dae.add_jac(Gy0, self.u, self.psiq, self.vd)

        dae.add_jac(Gy0, -1.0, self.psid, self.psid)
        dae.add_jac(Gy0, self.u, self.psid, self.vq)

        dae.add_jac(Gy0, self.u, self.Id, self.psid)

        dae.add_jac(Gy0, self.u, self.Iq, self.psiq)

        dae.add_jac(Fx0, -mul(self.u, self.iM, self.D) + 1e-6, self.omega,
                    self.omega)
        dae.add_jac(Fy0, mul(self.u, self.iM), self.omega, self.pm)


class Flux2(object):
    """Full electromagnetic transient of flux linkage
       d(psid) = wb * (ra * id + omega * psiq + vd)
       d(psiq) = wb * (ra * iq - omega * psid + vq)
       """

    def __init__(self):
        self._states.extend(['psid', 'psiq'])
        self._fnamex.extend(['\\psi_d', '\\psi_q'])
        self._init()

    def init1(self, dae):
        dae.x[self.psiq] = -mul(self.ra, dae.y[self.Id]) - dae.y[self.vd]
        dae.x[self.psid] = mul(self.ra, dae.y[self.Iq]) + dae.y[self.vq]

    def fcall(self, dae):
        wn = mul(self.system.wb, self.u)
        dae.f[self.psid] = mul(
            wn,
            mul(self.ra, dae.y[self.Id]) + mul(
                dae.x[self.omega], dae.x[self.psiq]) + dae.y[self.vd])
        dae.f[self.psiq] = mul(
            wn,
            mul(self.ra, dae.y[self.Iq]) - mul(
                dae.x[self.omega], dae.x[self.psid]) + dae.y[self.vq])

    def fxcall(self, dae):
        wn = mul(self.system.wb, self.u)
        dae.add_jac(Fy, mul(wn, self.ra), self.psid, self.Id)
        dae.add_jac(Fx, mul(wn, dae.x[self.omega]), self.psid, self.psiq)
        dae.add_jac(Fx, mul(wn, dae.x[self.psiq]), self.psid, self.omega)

        dae.add_jac(Fy, mul(wn, self.ra), self.psiq, self.Iq)
        dae.add_jac(Fx, -mul(wn, dae.x[self.omega]), self.psiq, self.psid)
        dae.add_jac(Fx, -mul(wn, dae.x[self.psid]), self.psiq, self.omega)

    def jac0(self, dae):
        wn = mul(self.system.wb, self.u)
        dae.add_jac(Fy0, wn, self.psid, self.vd)

        dae.add_jac(Fy0, wn, self.psiq, self.vq)


class Syn2(Ord2, Flux0):
    """2nd-order generator model. Inherited from (Ord2, Flux0)  """

    def __init__(self, system, name):
        Ord2.__init__(self, system, name)
        Flux0.__init__(self)

    def init1(self, dae):
        Ord2.init1(self, dae)
        Flux0.init1(self, dae)

    def gcall(self, dae):
        Ord2.gcall(self, dae)
        Flux0.gcall(self, dae)

    def fcall(self, dae):
        Ord2.fcall(self, dae)
        Flux0.fcall(self, dae)

    def jac0(self, dae):
        Ord2.jac0(self, dae)
        Flux0.jac0(self, dae)

    def gycall(self, dae):
        Ord2.gycall(self, dae)
        Flux0.gycall(self, dae)

    def fxcall(self, dae):
        Ord2.fxcall(self, dae)
        Flux0.fxcall(self, dae)


class Syn6a(Ord6a, Flux0):
    """GENROU, 6th-order generator model with xd2 = xq2."""

    def __init__(self, system, name):
        Ord6a.__init__(self, system, name)
        Flux0.__init__(self)

    def init1(self, dae):
        Ord6a.init1(self, dae)
        Flux0.init1(self, dae)

    def gcall(self, dae):
        Ord6a.gcall(self, dae)
        Flux0.gcall(self, dae)

    def fcall(self, dae):
        Ord6a.fcall(self, dae)
        Flux0.fcall(self, dae)

    def jac0(self, dae):
        Ord6a.jac0(self, dae)
        Flux0.jac0(self, dae)

    def gycall(self, dae):
        Ord6a.gycall(self, dae)
        Flux0.gycall(self, dae)

    def fxcall(self, dae):
        Ord2.fxcall(self, dae)
        Flux0.fxcall(self, dae)



# ANDES, a power system simulation tool for research.
#
# Copyright 2015-2017 Hantao Cui
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Power system class
"""

import configparser
import importlib
import logging
import os
from operator import itemgetter
from collections import OrderedDict

from . import routines
from .config import System
from .consts import pi, rad2deg
from .models import non_jits, jits, JIT, all_models_list
from .utils import get_config_load_path
from .variables import FileMan, DevMan, DAE, VarName, VarOut, Call, Report

try:
    from andes_addon.streaming import Streaming
    STREAMING = True
except ImportError:
    STREAMING = False

logger = logging.getLogger(__name__)


class PowerSystem(object):
    """
    A power system class to hold models, routines, DAE numerical values,
    file manager, call manager, variable names and valuable values.
    """

    def __init__(self,
                 case='',
                 pid=-1,
                 no_output=False,
                 dump_raw=None,
                 output=None,
                 dynfile=None,
                 addfile=None,
                 config=None,
                 input_path=None,
                 input_format=None,
                 output_format=None,
                 output_path='',
                 gis=None,
                 dime=None,
                 tf=None,
                 **kwargs):
        """
        PowerSystem constructor

        Parameters
        ----------
        case : str, optional
            Path to the case file

        pid : idx, optional
            Process index

        no_output : bool, optional
            ``True`` to disable all updates
        dump_raw : None or str, optional
            Path to the file to dump the raw parameters in the dm format

        output : None or str, optional
            Output case file name, NOT implemented yet

        dynfile : None or str, optional
            Path to the dynamic file for some formats, for example, ``dyr``
            for PSS/E

        addfile : None or str, optional
            Path to the additional dynamic file in the dm format

        config : None or str, optional
            Path to the andes.conf file

        input_format : None or str, optional
            Suggested input file format

        output_format : None or str, optional
            Requested dump file format, NOT implemented yet

        gis : None or str, optional
            Path to the GIS file

        dime : None or str, optional
            DiME server address

        tf : None or float, optional
            Time-domain simulation end time

        kwargs : dict, optional
            Other keyword args
        """
        # set internal flags based on the arguments
        self.pid = pid
        self.files = FileMan(case,
                             input_format=input_format,
                             addfile=addfile,
                             config=config,
                             input_path=input_path,
                             no_output=no_output,
                             dynfile=dynfile,
                             dump_raw=dump_raw,
                             output_path=output_path,
                             output_format=output_format,
                             output=output,
                             **kwargs)

        self.config = System()

        conf_path = get_config_load_path(self.files.config)
        self.load_config(conf_path=conf_path, sys_only=True)

        self.routine_import()
        self.load_config(conf_path=conf_path, routine_only=True)

        self.devman = DevMan(self)
        self.call = Call(self)
        self.dae = DAE(self)
        self.varname = VarName(self)
        self.varout = VarOut(self)
        self.report = Report(self)

        self.loaded_groups = []

        if dime:
            self.config.dime_enable = True
            self.config.dime_server = dime

        if tf:
            self.tds.config.tf = tf

        self.streaming = None
        if not STREAMING:
            if self.config.dime_enable:
                self.config.dime_enable = False
                logger.warning('Missing andes_addon for DiME streaming.')
        else:
            if self.config.dime_enable:
                logger.info('Connecting to DiME at {}'.format(self.config.dime_server))
                self.streaming = Streaming(self)

        self.model_import()

    @property
    def freq(self):
        """System base frequency"""
        return self.config.freq

    @freq.setter
    def freq(self, freq):
        if freq <= 0:
            self.config.freq = 1
        else:
            self.config.freq = freq

    @property
    def wb(self):
        """System base radial frequency"""
        return 2 * pi * self.config.freq

    @property
    def mva(self):
        """System base power in mega voltage-ampere"""
        return self.config.mva

    @mva.setter
    def mva(self, mva):
        self.config.mva = mva

    def setup(self):
        """
        Set up the power system object by executing the following workflow:

         * Sort the loaded models to meet the initialization sequence
         * Create call strings for routines
         * Call the ``setup`` function of the loaded models
         * Assign addresses for the loaded models
         * Call ``dae.setup`` to assign memory for the numerical dae structure
         * Convert model parameters to the system base

        Returns
        -------
        PowerSystem
            The instance of the PowerSystem
        """

        self.devman.sort_device()
        self.call.setup()
        self.model_setup()
        self.xy_addr0()
        self.dae.setup()
        self.to_sysbase()

        return self

    def to_sysbase(self):
        """
        Convert model parameters to system base. This function calls the
        ``data_to_sys_base`` function of the loaded models.

        Returns
        -------
        None
        """
        if self.config.base:
            for item in self.devman.devices:
                self.__dict__[item].data_to_sys_base()

    def to_elembase(self):
        """
        Convert parameters back to element base. This function calls the
        ```data_to_elem_base``` function.

        Returns
        -------
        None
        """
        if self.config.base:
            for item in self.devman.devices:
                self.__dict__[item].data_to_elem_base()

    def group_add(self, name='Ungrouped'):
        """
        Dynamically add a group instance to the system if not exist.

        Parameters
        ----------
        name : str, optional ('Ungrouped' as default)
            Name of the group

        Returns
        -------
        None
        """
        if not hasattr(self, name):
            self.__dict__[name] = Group(self, name)
            self.loaded_groups.append(name)

    def model_import(self):
        """
        Import and instantiate the non-JIT models and the JIT models.

        Models defined in ``jits`` and ``non_jits`` in ``models/__init__.py``
        will be imported and instantiated accordingly.

        Returns
        -------
        None
        """
        # non-JIT models
        for file, pair in non_jits.items():
            for cls, name in pair.items():
                themodel = importlib.import_module('andes.models.' + file)
                theclass = getattr(themodel, cls)
                self.__dict__[name] = theclass(self, name)

                group = self.__dict__[name]._group
                if group is None:
                    raise ValueError("Class definition incomplete. Group not defined for class {}".format(name))
                self.group_add(group)
                self.__dict__[group].register_model(name)

                self.devman.register_device(name)

        # import JIT models
        for file, pair in jits.items():
            for cls, name in pair.items():
                self.__dict__[name] = JIT(self, file, cls, name)

    def routine_import(self):
        """
        Dynamically import routines as defined in ``routines/__init__.py``.

        The command-line argument ``--routine`` is defined in ``__cli__`` in
        each routine file. A routine instance will be stored in the system
        instance with the name being all lower case.

        For example, a routine for power flow study should be defined in
        ``routines/pflow.py`` where ``__cli__ = 'pflow'``. The class name
        for the routine should be ``Pflow``. The routine instance will be
        saved to ``PowerSystem.pflow``.

        Returns
        -------
        None
        """
        for r in routines.__all__:
            file = importlib.import_module('.' + r.lower(), 'andes.routines')
            self.__dict__[r.lower()] = getattr(file, r)(self)

    def model_setup(self):
        """
        Call the ``setup`` function of the loaded models. This function is
        to be called after parsing all the data files during the system set up.

        Returns
        -------
        None
        """
        for device in self.devman.devices:
            try:
                self.__dict__[device].setup()
            except Exception as e:
                raise e

    def xy_addr0(self):
        """
        Assign indicies and variable names for variables used in power flow

        For each loaded model with the ``pflow`` flag as ``True``, the following
        functions are called sequentially:

         * ``_addr()``
         * ``_intf_network()``
         * ``_intf_ctrl()``

        After resizing the ``varname`` instance, variable names from models
        are stored by calling ``_varname()``

        Returns
        -------
        None
        """
        for device, pflow in zip(self.devman.devices, self.call.pflow):
            if pflow:
                self.__dict__[device]._addr()
                self.__dict__[device]._intf_network()
                self.__dict__[device]._intf_ctrl()

        self.varname.resize()

        for device, pflow in zip(self.devman.devices, self.call.pflow):
            if pflow:
                self.__dict__[device]._varname()

    def xy_addr1(self):
        """
        Assign indices and variable names for variables after power flow.
        This function is for loaded models that do not have the ``pflow``
        flag as ``True``.
        """
        for device, pflow in zip(self.devman.devices, self.call.pflow):
            if not pflow:
                self.__dict__[device]._addr()
                self.__dict__[device]._intf_network()
                self.__dict__[device]._intf_ctrl()

        self.varname.resize()

        for device, pflow in zip(self.devman.devices, self.call.pflow):
            if not pflow:
                self.__dict__[device]._varname()

    def rmgen(self, idx):
        """
        Remove the static generators if their dynamic models exist

        Parameters
        ----------
        idx : list
            A list of static generator idx
        Returns
        -------
        None
        """
        stagens = []
        for device, stagen in zip(self.devman.devices, self.call.stagen):
            if stagen:
                stagens.append(device)
        for gen in idx:
            for stagen in stagens:
                if gen in self.__dict__[stagen].uid.keys():
                    self.__dict__[stagen].disable_gen(gen)

    def check_event(self, sim_time):
        """
        Check for event occurrance for``Event`` group models at ``sim_time``

        Parameters
        ----------
        sim_time : float
            The current simulation time

        Returns
        -------
        list
            A list of model names who report (an) event(s) at ``sim_time``
        """
        ret = []
        for model in self.__dict__['Event'].all_models:
            if self.__dict__[model].is_time(sim_time):
                ret.append(model)

        if self.Breaker.is_time(sim_time):
            ret.append('Breaker')

        return ret

    def get_data_example(self, model, n_per_line=5):
        """
        Return a string of example data entry that can be used in a dm input file

        Returns
        -------
        str
            A string containing the example data
        """
        if model not in all_models_list:
            raise KeyError('Model <{}> is invalid'.format(model))

        model_name = self.__dict__[model]._name

        data = OrderedDict()
        data.update({'idx': None, 'name': '"{}"'.format(model_name)})
        data.update(self.__dict__[model]._data)

        out = '{}, '.format(model_name)
        nspace = len(out)
        n_param = len(data)

        for idx, (key, val) in enumerate(data.items()):
            if (idx > 0) and (divmod(idx, n_per_line)[1] == 0):
                out += '\n' + ' ' * nspace
            if idx == n_param - 1:
                out += '{} = {}'.format(key, val)
            else:
                out += '{} = {}, '.format(key, val)

        return out

    def get_event_times(self):
        """
        Return event times of Fault, Breaker and other timed events

        Returns
        -------
        list
            A sorted list of event times
        """
        times = []

        times.extend(self.Breaker.get_times())

        for model in self.__dict__['Event'].all_models:
            times.extend(self.__dict__[model].get_times())

        if times:
            times = sorted(list(set(times)))

        return times

    def load_config(self, conf_path, sys_only=False, routine_only=False):
        """
        Load config from an ``andes.conf`` file.

        This function creates a ``configparser.ConfigParser`` object to read
        the specified conf file and calls the ``load_config`` function of the
        config instances of the system and the routines.

        Parameters
        ----------
        conf_path : None or str
            Path to the Andes config file. If ``None``, the function body
            will not run.

        Returns
        -------
        None
        """
        if conf_path is None:
            return

        conf = configparser.ConfigParser()
        conf.read(conf_path)
        if sys_only:
            self.config.load_config(conf)
            return
        if routine_only:
            for r in routines.__all__:
                self.__dict__[r.lower()].config.load_config(conf)

        logger.debug('Loaded config file from {}.'.format(conf_path))

    def dump_config(self, file_path):
        """
        Dump system and routine configurations to an rc-formatted file.

        Parameters
        ----------
        file_path : str
            path to the configuration file. The user will be prompted if the
            file already exists.

        Returns
        -------
        None
        """
        if os.path.isfile(file_path):
            logger.debug('File {} alreay exist. Overwrite? [y/N]'.format(file_path))
            choice = input('File {} alreay exist. Overwrite? [y/N]'.format(file_path)).lower()
            if len(choice) == 0 or choice[0] != 'y':
                logger.info('File not overwritten.')
                return

        conf = self.config.dump_conf()
        for r in routines.__all__:
            conf = self.__dict__[r.lower()].config.dump_conf(conf)

        with open(file_path, 'w') as f:
            conf.write(f)

        logger.info('Config written to {}'.format(file_path))

    def check_islands(self, show_info=False):
        """
        Check the connectivity for the ac system

        Parameters
        ----------
        show_info : bool
            Show information when the system has islands. To be used when
            initializing power flow.

        Returns
        -------
        None
        """
        if not hasattr(self, 'Line'):
            logger.error('<Line> device not found.')
            return
        self.Line.connectivity(self.Bus)

        if show_info is True:

            if len(self.Bus.islanded_buses) == 0 and len(
                    self.Bus.island_sets) == 0:
                logger.debug('System is interconnected.')
            else:
                logger.info(
                    'System contains {:d} islands and {:d} islanded buses.'.
                    format(
                        len(self.Bus.island_sets),
                        len(self.Bus.islanded_buses)))

            nosw_island = []  # no slack bus island
            msw_island = []  # multiple slack bus island
            for idx, island in enumerate(self.Bus.island_sets):
                nosw = 1
                for item in self.SW.bus:
                    if self.Bus.uid[item] in island:
                        nosw -= 1
                if nosw == 1:
                    nosw_island.append(idx)
                elif nosw < 0:
                    msw_island.append(idx)

            if nosw_island:
                logger.warning(
                    'Slack bus is not defined for {:g} island(s).'.format(
                        len(nosw_island)))
            if msw_island:
                logger.warning(
                    'Multiple slack buses are defined for {:g} island(s).'.
                    format(len(nosw_island)))

            if (not nosw_island) and (not msw_island):
                logger.debug(
                    'Each island has a slack bus correctly defined.')

    def get_busdata(self, sort_names=False):
        """
        get ac bus data from solved power flow
        """
        if self.pflow.solved is False:
            logger.error('Power flow not solved when getting bus data.')
            return tuple([False] * 8)
        idx = self.Bus.idx
        names = self.Bus.name
        Vm = [self.dae.y[x] for x in self.Bus.v]
        if self.pflow.config.usedegree:
            Va = [self.dae.y[x] * rad2deg for x in self.Bus.a]
        else:
            Va = [self.dae.y[x] for x in self.Bus.a]

        Pg = [self.Bus.Pg[x] for x in range(self.Bus.n)]
        Qg = [self.Bus.Qg[x] for x in range(self.Bus.n)]
        Pl = [self.Bus.Pl[x] for x in range(self.Bus.n)]
        Ql = [self.Bus.Ql[x] for x in range(self.Bus.n)]

        if sort_names:
            ret = (list(x) for x in zip(*sorted(
                zip(idx, names, Vm, Va, Pg, Qg, Pl, Ql), key=itemgetter(0))))
        else:
            ret = idx, names, Vm, Va, Pg, Qg, Pl, Ql

        return ret

    def get_nodedata(self, sort_names=False):
        """
        get dc node data from solved power flow
        """
        if not self.Node.n:
            return
        if not self.pflow.solved:
            logger.error('Power flow not solved when getting bus data.')
            return tuple([False] * 7)
        idx = self.Node.idx
        names = self.Node.name
        V = [self.dae.y[x] for x in self.Node.v]

        if sort_names:
            ret = (list(x)
                   for x in zip(*sorted(zip(idx, names, V), key=itemgetter(0))))
        else:
            ret = idx, names, V

        return ret

    def get_linedata(self, sort_names=False):
        """
        get line data from solved power flow
        """
        if not self.pflow.solved:
            logger.error('Power flow not solved when getting line data.')
            return tuple([False] * 7)
        idx = self.Line.idx
        fr = self.Line.bus1
        to = self.Line.bus2

        Sloss = self.Line.S1 + self.Line.S2

        Pfr = list(self.Line.S1.real())
        Qfr = list(self.Line.S1.imag())
        Pto = list(self.Line.S2.real())
        Qto = list(self.Line.S2.imag())

        Ploss = list(Sloss.real())
        Qloss = list(Sloss.imag())

        if sort_names:
            ret = (list(x) for x in zip(*sorted(
                zip(idx, fr, to, Pfr, Qfr, Pto, Qto, Ploss, Qloss),
                key=itemgetter(0))))
        else:
            ret = idx, fr, to, Pfr, Qfr, Pto, Qto, Ploss, Qloss

        return ret


class GroupMeta(type):
    def __new__(cls, name, base, attr_dict):
        return super(GroupMeta, cls).__new__(cls, name, base, attr_dict)


class Group(metaclass=GroupMeta):
    """
    Group class for registering models and elements.

    Also handles reading and setting attributes.
    """

    def __init__(self, system, name):
        self.system = system
        self.name = name
        self.all_models = []
        self._idx_model = {}
        self._idx = []

    def register_model(self, model):
        """
        Register ``model`` to this group

        :param model: model name
        :return: None
        """

        if not isinstance(model, str):
            # TODO: consider removing this constrain
            raise KeyError("Name of the model must be a string")
        if model not in self.all_models:
            self.all_models.append(model)

    def register_element(self, model, idx):
        """
        Register element with index ``idx`` to ``model``

        :param model: model name
        :param idx: element idx
        :return: final element idx
        """

        if idx is None:
            idx = model + '_' + str(len(self._idx_model))

        self._idx_model[idx] = model
        self._idx.append(idx)

        return idx

    def get_field(self, field, idx):
        """
        Return the field ``field`` of elements ``idx`` in the group

        :param field: field name
        :param idx: element idx
        :return: values of the requested field
        """
        ret = []
        scalar = False

        # TODO: ensure idx is unique in this Group

        if isinstance(idx, (int, float, str)):
            scalar = True
            idx = [idx]

        models = [self._idx_model[i] for i in idx]

        for i, m in zip(idx, models):
            ret.append(self.system.__dict__[m].get_field(field, idx=i))

        if scalar is True:
            return ret[0]
        else:
            return ret

    def set_field(self, field, idx, value):
        """
        Set the field ``field`` of elements ``idx`` to ``value``.

        This function does not if the field is valid for all models.

        :param field: field name
        :param idx: element idx
        :param value: value of fields to set
        :return: None
        """

        if isinstance(idx, (int, float, str)):
            idx = [idx]
        if isinstance(value, (int, float)):
            value = [value]

        models = [self._idx_model[i] for i in idx]

        for i, m, v in zip(idx, models, value):
            if not hasattr(self.system.__dict__[m], field):
                raise KeyError('Field <{}> is not valid for model <{}>'.format(field, m))

            uid = self.system.__dict__[m].get_uid(idx)
            self.system.__dict__[m].__dict__[field][uid] = v



from cvxopt import mul, div  # NOQA

from .base import ModelBase
from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA
from ..utils.math import zeros, ones


class BusFreq(ModelBase):
    """Bus frequency estimation based on angle derivative"""

    def __init__(self, system, name):
        super(BusFreq, self).__init__(system, name)
        self._group = 'Measurement'
        self._name = 'BusFreq'
        self.param_remove('Vn')
        self._data.update({
            'bus': None,
            'Tf': 0.1,
            'Tw': 1.0,
            'Td': 0.5,
        })
        self._params.extend(['Tf', 'Tw', 'Td'])
        self._descr.update({
            'bus':
            'bus idx',
            'Tf':
            'low-pass filter time constant',
            'Tw':
            'washout filter time constant for phase angle',
            'Td':
            'washout filter time constant for frequency',
        })
        self._units.update({
            'bus': 'na',
            'Tf': 'sec',
            'Tw': 'sec',
        })
        self._mandatory.extend(['bus'])
        self._states.extend(['xt', 'w', 'xdw'])
        self._fnamex.extend(['x_\\theta', '\\omega', 'x_\\omega'])
        self._algebs.extend(['dwdt'])
        self._fnamey.extend(['\\frac{d\\omega}{dt}'])

        self.calls.update({
            'init1': True,
            'fcall': True,
            'gcall': True,
            'jac0': True,
        })
        self._service.extend(['iTf', 'iTw', 'iwn', 'a0', 'dw', 'iTd'])
        self._init()

    def init1(self, dae):
        self.copy_data_ext(model='Bus', field='a', dest='a', idx=self.bus)
        self.iTf = div(self.u, self.Tf)
        self.iTd = div(self.u, self.Td)
        self.iTw = div(self.u, self.Tw)
        self.iwn = div(self.u, self.system.wb)
        self.a0 = dae.y[self.a]

        dae.x[self.xt] = zeros(self.n, 1)
        dae.x[self.w] = ones(self.n, 1)
        # dae.y[self.dw] = zeros(self.n, 1)
        dae.x[self.xdw] = self.iTd

    def fcall(self, dae):
        self.dw = -dae.x[self.xt] + mul(self.iwn, self.iTf,
                                        dae.y[self.a] - self.a0)
        dae.f[self.xt] = mul(
            self.iTf,
            mul(self.iwn, self.iTf, dae.y[self.a] - self.a0) - dae.x[self.xt])
        dae.f[self.w] = mul(self.dw + 1 - dae.x[self.w], self.iTw)
        dae.f[self.xdw] = mul(self.iTd,
                              mul(self.iTd, dae.x[self.w]) - dae.x[self.xdw])

    def gcall(self, dae):
        dae.g[self.dwdt] = (
            mul(self.iTd, dae.x[self.w]) - dae.x[self.xdw]) - dae.y[self.dwdt]

    def jac0(self, dae):
        dae.add_jac(Fy0, mul(self.iTf**2, self.iwn), self.xt, self.a)
        dae.add_jac(Fx0, -self.iTf, self.xt, self.xt)

        dae.add_jac(Fx0, -self.iTw, self.w, self.w)

        dae.add_jac(Fx0, -self.iTw, self.w, self.xt)
        dae.add_jac(Fy0, mul(self.iTw, self.iwn, self.iTf), self.w, self.a)

        dae.add_jac(Gx0, self.iTd, self.dwdt, self.w)
        dae.add_jac(Gx0, -1, self.dwdt, self.xdw)

        dae.add_jac(Fx0, self.iTd**2, self.xdw, self.w)
        dae.add_jac(Fx0, -self.iTd, self.xdw, self.xdw)

        dae.add_jac(Gy0, -1, self.dwdt, self.dwdt)


class PMU(ModelBase):
    """Phasor measurement unit described by low-pass filters"""

    def __init__(self, system, name):
        super(PMU, self).__init__(system, name)
        self._group = 'Measurement'
        self._data.update({
            'Tv': 0.1,
            'Ta': 0.1,
            'fn': 60,
            'bus': None,
        })
        self._params.extend(['Tv', 'Ta', 'fn'])
        self._descr.update({
            'Tv': 'Voltage magnitude time constant',
            'Ta': 'Voltage angle time constant',
            'fn': 'Frequency base',
            'bus': 'Bus idx'
        })
        self._states.extend(['vm', 'am'])
        self._fnamex.extend(['V_m', '\\theta_m'])
        self._mandatory.extend(['bus'])
        self._service.extend(['iTv', 'iTa'])
        self._init()
        self.calls.update({
            'fcall': True,
            'jac0': True,
            'init1': True,
        })

    def init1(self, dae):
        self.copy_data_ext(model='Bus', field='a', idx=self.bus)
        self.copy_data_ext(model='Bus', field='v', idx=self.bus)
        self.iTv = div(self.u, self.Tv)
        self.iTa = div(self.u, self.Ta)
        dae.x[self.am] = dae.y[self.a]
        dae.x[self.vm] = dae.y[self.v]

    def fcall(self, dae):
        dae.f[self.vm] = mul(dae.y[self.v] - dae.x[self.vm], self.iTv)
        dae.f[self.am] = mul(dae.y[self.a] - dae.x[self.am], self.iTa)

    def jac0(self, dae):
        dae.add_jac(Fy0, self.iTv, self.vm, self.v)
        dae.add_jac(Fx0, -self.iTv + 1e-6, self.vm, self.vm)

        dae.add_jac(Fy0, self.iTa, self.am, self.a)
        dae.add_jac(Fx0, -self.iTa + 1e-6, self.am, self.am)



from .cached import cached  # NOQA
from .misc import get_config_load_path  # NOQA
from .stock_case import get_stock_case  # NOQA
from .tab import Tab  # NOQA
from .time import elapsed  # NOQA

__all__ = [
    'cached',
    'tab',
    'math',
    'time',
    'misc',
    'stock_case',
    'get_config_load_path'
]



from . import ConfigBase
from ..utils.cached import cached


class Eig(ConfigBase):
    def __init__(self, **kwargs):
        self.neig = 1
        self.method = 1
        self.map = 1
        self.matrix = 4
        self.report = ''
        self.eigs = ''
        self.pf = ''
        self.plot = False
        super(Eig, self).__init__(**kwargs)

    @cached
    def config_descr(self):
        descriptions = {}
        return descriptions



from .base import ConfigBase  # NOQA
from .cpf import CPF  # NOQA
from .eig import Eig  # NOQA
from .pflow import Pflow  # NOQA
from .system import System  # NOQA
from .tds import Tds  # NOQA

__all__ = [
    'system',
    'pflow',
    'tds',
    'eig',
    'cpf',
]



from cvxopt import mul, div  # NOQA

from .base import ModelBase
from ..consts import Fx0, Fy0, Gx0, Gy0  # NOQA


class COI(ModelBase):
    def __init__(self, system, name):
        super(COI, self).__init__(system, name)
        self._group = 'Calculation'
        self._data.update({
            'syn': None,
            'Td': 1,
        })
        self._descr.update({
            'syn': 'synchronous generator list',
            'Td': 'washout filter time constant',
        })
        self._params.extend({'Td'})
        self._algebs.extend(['delta', 'omega', 'dwdt'])
        self._states.extend(['xdw'])
        self.calls.update({
            'init1': True,
            'gcall': True,
            'fcall': True,
            'jac0': True,
        })
        self._service.extend(
            ['H', 'M', 'Mtot', 'usyn', 'gdelta', 'gomega', 'iTd'])
        self._fnamey.extend(['\\delta', '\\omega', '\\frac{d\\omega}{dt}'])
        self._fnamex.extend(['x\\omega'])
        self._init()

    def init1(self, dae):
        for item in self._service:
            self.__dict__[item] = [[]] * self.n

        dae.y[self.omega] = 1
        for idx, item in enumerate(self.syn):
            self.usyn[idx] = self.read_data_ext(
                'Synchronous', field='u', idx=item)
            self.M[idx] = self.read_data_ext(
                'Synchronous', field='M', idx=item)
            self.Mtot[idx] = sum(self.M[idx])
            self.gdelta[idx] = self.read_data_ext(
                'Synchronous', field='delta', idx=item)
            self.gomega[idx] = self.read_data_ext(
                'Synchronous', field='omega', idx=item)

            dae.y[self.delta[idx]] = sum(
                mul(self.M[idx], dae.x[self.gdelta[idx]])) / self.Mtot[idx]

        self.iTd = div(self.u, self.Td)
        dae.x[self.xdw] = self.iTd

    def gcall(self, dae):
        for idx in range(self.n):
            dae.g[self.omega[idx]] = dae.y[self.omega[idx]] - sum(
                mul(self.M[idx], dae.x[self.gomega[idx]])) / self.Mtot[idx]
            dae.g[self.delta[idx]] = dae.y[self.delta[idx]] - sum(
                mul(self.M[idx], dae.x[self.gdelta[idx]])) / self.Mtot[idx]
        dae.g[self.dwdt] = (mul(self.iTd, dae.y[self.omega]) -
                            dae.x[self.xdw]) - dae.y[self.dwdt]

    def jac0(self, dae):
        dae.add_jac(Gy0, 1, self.omega, self.omega)
        dae.add_jac(Gy0, 1, self.delta, self.delta)

        dae.add_jac(Gy0, self.iTd, self.dwdt, self.omega)
        dae.add_jac(Gx0, -1, self.dwdt, self.xdw)

        dae.add_jac(Fy0, self.iTd**2, self.xdw, self.omega)
        dae.add_jac(Fx0, -self.iTd, self.xdw, self.xdw)

        dae.add_jac(Gy0, -1, self.dwdt, self.dwdt)

    def fcall(self, dae):
        dae.f[self.xdw] = mul(
            self.iTd,
            mul(self.iTd, dae.y[self.omega]) - dae.x[self.xdw])



import importlib
import logging
from time import strftime

from cvxopt import mul

from andes import __version__ as version
from ..formats import all_formats
from ..utils import elapsed

logger = logging.getLogger(__name__)


class Report(object):
    """
    Report class to store system static analysis reports
    """

    def __init__(self, system):
        self.system = system
        self.basic = {}
        self.extended = {}

        self._basic = [
            'nbus', 'ngen', 'ngen_on', 'nload', 'nshunt', 'nline', 'ntransf',
            'narea'
        ]
        self._basic_name = [
            'Buses', 'Generators', 'Committed Gens', 'Loads', 'Shunts',
            'Lines', 'Transformers', 'Areas'
        ]

        self._extended = [
            'Ptot', 'Pon', 'Pg', 'Qtot_min', 'Qtot_max', 'Qon_min', 'Qon_max',
            'Qg', 'Pl', 'Ql', 'Psh', 'Qsh', 'Ploss', 'Qloss', 'Pch', 'Qch'
        ]

        for item in self._basic:
            self.basic[item] = 0.0
        for item in self._extended:
            self.extended[item] = 0.0

    @property
    def info(self):
        info = list()
        info.append('ANDES' + ' ' + version + '\n')
        info.append('Copyright (C) 2015-2018 Hantao Cui\n\n')
        info.append('ANDES comes with ABSOLUTELY NO WARRANTY\n')
        info.append('Use this software AT YOUR OWN RISK\n\n')
        info.append('Case file: ' + self.system.files.case + '\n')
        info.append('report Time: ' + strftime("%m/%d/%Y %I:%M:%S %p") +
                    '\n\n')
        if self.system.pflow.solved is True:
            info.append('Power flow method: ' +
                        self.system.pflow.config.method.upper() + '\n')
            info.append('Number of iterations: ' + str(
                self.system.pflow.niter) +
                        '\n')
            info.append('Flat-start: ' +
                        ('Yes' if self.system.pflow.config.flatstart else 'No') + '\n')

        return info

    def _update_summary(self, system):
        """
        Update the summary data

        Parameters
        ----------
        system

        Returns
        -------
        None
        """
        self.basic.update({
            'nbus': system.Bus.n,
            'ngen': system.PV.n + system.SW.n,
            'ngen_on': sum(system.PV.u) + sum(system.SW.u),
            'nload': system.PQ.n,
            'nshunt': system.Shunt.n,
            'nline': system.Line.n,
            'ntransf': system.Line.trasf.count(True),
            'narea': system.Area.n,
        })

    def _update_extended(self, system):
        """Update the extended data"""
        if self.system.pflow.solved is False:
            logger.warning(
                'Cannot update extended summary. Power flow not solved.')
            return

        Sloss = sum(system.Line.S1 + system.Line.S2)
        self.extended.update({
            'Ptot':
            sum(system.PV.pmax) + sum(system.SW.pmax),
            'Pon':
            sum(mul(system.PV.u, system.PV.pmax)),
            'Pg':
            sum(system.Bus.Pg),
            'Qtot_min':
            sum(system.PV.qmin) + sum(system.SW.qmin),
            'Qtot_max':
            sum(system.PV.qmax) + sum(system.SW.qmax),
            'Qon_min':
            sum(mul(system.PV.u, system.PV.qmin)),
            'Qon_max':
            sum(mul(system.PV.u, system.PV.qmax)),
            'Qg':
            round(sum(system.Bus.Qg), 6),
            'Pl':
            round(sum(system.PQ.p), 6),
            'Ql':
            round(sum(system.PQ.q), 6),
            'Psh':
            0.0,
            'Qsh':
            round(sum(system.PQ.q) - sum(system.Bus.Ql), 6),
            'Ploss':
            round(Sloss.real, 6),
            'Qloss':
            round(Sloss.imag, 6),
            'Pch':
            round(sum(system.Line.Pchg1 + system.Line.Pchg2), 6),
            'Qch':
            round(sum(system.Line.Qchg1 + system.Line.Qchg2), 6),
        })

    def update(self, content=None):
        """
        Update values based on the requested content

        Parameters
        ----------
        content

        Returns
        -------

        """
        if not content:
            return
        if content == 'summary' or 'extended' or 'powerflow':
            self._update_summary(self.system)
        if content == 'extended' or 'powerflow':
            self._update_extended(self.system)

    def write(self, content=None):
        """
        Write report to file.

        Parameters
        ----------
        content: str
            'summary', 'extended', 'powerflow'

        """
        if self.system.files.no_output is True:
            return

        t, _ = elapsed()

        if not content:
            logger.warning('report content not specified.')
            return

        self.update(content)

        system = self.system
        file = system.files.output
        export = all_formats.get(system.config.export, 'txt')
        module = importlib.import_module('andes.formats.' + export)
        dump_data = getattr(module, 'dump_data')

        text = list()
        header = list()
        rowname = list()
        data = list()

        text.append(self.info)
        header.append(None)
        rowname.append(None)
        data.append(None)

        if content == 'summary' or 'extended' or 'powerflow':
            text.append(['SUMMARY:\n'])
            header.append(None)
            rowname.append(self._basic_name)
            data.append([self.basic[item] for item in self._basic])

        if content == 'extended' or 'powerflow':
            text.append(['EXTENDED SUMMARY:\n'])
            header.append(['P (pu)', 'Q (pu)'])
            rowname.append(
                ['Generation', 'Load', 'Shunt Inj', 'Losses', 'Line Charging'])
            Pcol = [
                self.extended['Pg'],
                self.extended['Pl'],
                self.extended['Psh'],
                self.extended['Ploss'],
                self.extended['Pch'],
            ]

            Qcol = [
                self.extended['Qg'],
                self.extended['Ql'],
                self.extended['Qsh'],
                self.extended['Qloss'],
                self.extended['Qch'],
            ]

            data.append([Pcol, Qcol])

        if content == 'powerflow' and system.pflow.solved:
            idx, name, Vm, Va, Pg, Qg, Pl, Ql = system.get_busdata()
            Va_unit = 'deg' if system.pflow.config.usedegree else 'rad'
            text.append(['BUS DATA:\n'])
            # todo: consider system.pflow.config.units
            header.append([
                'Vm(pu)', 'Va({:s})'.format(Va_unit), 'Pg (pu)', 'Qg (pu)',
                'Pl (pu)', 'Ql (pu)'
            ])
            name = [str(i) + '-' + j[:8] for i, j in zip(idx, name)]
            rowname.append(name)
            data.append([Vm, Va, Pg, Qg, Pl, Ql])

            # Node data
            if hasattr(system, 'Node') and system.Node.n:
                idx, name, V = system.get_nodedata()
                text.append(['NODE DATA:\n'])
                header.append(['V(pu)'])
                rowname.append(name)
                data.append([V])

            # Line data
            name, fr, to, Pfr, Qfr, Pto, Qto, Ploss, Qloss = \
                system.get_linedata()
            text.append(['LINE DATA:\n'])
            header.append([
                'From Bus', 'To Bus', 'P From (pu)', 'Q From (pu)',
                'P To (pu)', 'Q To(pu)', 'P Loss(pu)', 'Q Loss(pu)'
            ])
            rowname.append(name)
            data.append([fr, to, Pfr, Qfr, Pto, Qto, Ploss, Qloss])

            # Additional Algebraic data
            text.append(['OTHER ALGEBRAIC VARIABLES:\n'])
            header.append([''])
            rowname.append(
                system.varname.unamey[2 * system.Bus.n:system.dae.m])
            data.append([round(i, 6) for i in system.dae.y[2 * system.Bus.n:]])

            # Additional State variable data
            if system.dae.n:
                text.append(['OTHER STATE VARIABLES:\n'])
                header.append([''])
                rowname.append(system.varname.unamex[:])
                data.append([round(i, 6) for i in system.dae.x[:]])

        dump_data(text, header, rowname, data, file)

        _, s = elapsed(t)
        logger.info('Report saved to <{:s}> in {:s}.'.format(system.files.output, s))



import importlib
import logging

from . import ConfigBase
from ..utils.cached import cached

logger = logging.getLogger(__name__)
try:
    klu = importlib.import_module('cvxoptklu.klu')
except ImportError:
    klu = None


class System(ConfigBase):
    def __init__(self, **kwargs):
        self.freq = 60.0
        self.mva = 100.0
        self.distrsw = False
        self.sparselib = 'klu'
        self.sparselib_alt = ['klu', 'umfpack', 'spsolve', 'cupy']
        self.export = 'txt'
        self.export_alt = ['txt', 'latex']
        self.coi = False
        self.connectivity = False
        self.error = 1
        self.static = 0
        self.nseries = 0
        self.forcepq = False
        self.forcez = False
        self.base = True
        self.dime_enable = False
        self.dime_name = 'sim'
        self.dime_server = 'ipc:///tmp/dime'
        super(System, self).__init__(**kwargs)

    @cached
    def config_descr(self):
        descriptions = {
            'freq': 'system base frequency',
            'mva': 'system base MVA',
            'distrsw': 'use distributed slack bus mode',
            'sparselib': 'sparse matrix library name',
            'export': 'help documentation export format',
            'coi': 'using Center of Inertia',
            'connectivity': 'connectivity check during TDS',
            'forcepq': 'force to use constant PQ load',
            'forcez': 'force to convert load to impedance',
            'base': 'convert  parameters to the system base',
        }
        return descriptions

    def check(self):
        """
        Check config data consistency

        Returns
        -------

        """
        if self.sparselib not in self.sparselib_alt:
            logger.warning("Invalid sparse library <{}>".format(self.sparselib))
            self.sparselib = 'umfpack'

        if self.sparselib == 'klu' and (klu is None):
            logger.info("KLU not found. Install with \"pip install cvxoptklu\" ")
            self.sparselib = 'umfpack'

        elif self.sparselib == 'cupy':
            try:
                cupy_sparse = importlib.import_module('cupyx.scipy.sparse')
                cupy_solve = importlib.import_module('cupyx.scipy.sparse.linalg.solve')
                _ = getattr(cupy_sparse, 'csc_matrix')  # NOQA
                _ = getattr(cupy_solve, 'lsqr')  # NOQA
            except (ImportError, AttributeError):
                logger.info("CuPy not found. Fall back to UMFPACK.")
                self.sparselib = 'umfpack'

        return True



import logging
import os
import time

import numpy as np
from cvxopt import matrix

logger = logging.getLogger(__name__)


class VarOut(object):
    """
    Output variable value recorder

    TODO: merge in tds.py
    """

    def __init__(self, system):
        """Constructor of empty Varout object"""
        self.system = system
        self.epoch_t = []
        self.t = []
        self.k = []
        self.vars = []
        self.vars_array = None

        self._np_block_rows = 600
        self._np_block_shape = (0, 0)
        self.np_vars = np.ndarray(shape=(0, 0))
        self.np_t = np.ndarray(shape=(0,))
        self.np_epoch_t = np.ndarray(shape=(0,))
        self.np_k = np.ndarray(shape=(0,))

        self.np_nrows = 0
        self.np_ncols = 0

        self.dat = None
        self._mode = 'w'

        self._last_t = 0
        self._last_epoch_t = 0
        self._last_vars = []

    def store(self, t, step):
        """
        Record the state/algeb values at time t to self.vars
        """
        epoch_time = time.time()
        max_cache = int(self.system.tds.config.max_cache)
        if len(self.vars) >= max_cache > 0:
            self.dump()
            self.vars = list()
            self.t = list()
            self.epoch_t = list()
            self.k = list()
            logger.debug(
                'varout cache cleared at simulation t = {:g}.'.format(
                    self.system.dae.t))
            self._mode = 'a'

        var_data = matrix([self.system.dae.x, self.system.dae.y])

        # ===== This code block is deprecated =====
        self.t.append(t)
        self.epoch_t.append(epoch_time)
        self.k.append(step)
        self.vars.append(var_data)
        # =========================================

        # temporary storage
        self._last_t = t
        self._last_epoch_t = epoch_time
        self._last_vars = list(var_data)

        #

        # clear data cache if written to disk
        if self.np_nrows >= max_cache > 0:
            self.dump_np_vars()
            self.np_vars = np.zeros(self._np_block_shape)
            self.np_nrows = 0
            self.np_t = np.zeros((self._np_block_rows,))
            self.np_epoch_t = np.zeros((self._np_block_rows,))
            self.np_k = np.zeros((self._np_block_rows,))
            logger.debug(
                'np_vars cache cleared at simulation t = {:g}.'.format(
                    self.system.dae.t))
            self._mode = 'a'

        # initialize before first-time adding data
        if self.np_nrows == 0:
            self.np_ncols = len(var_data)
            self._np_block_shape = (self._np_block_rows, self.np_ncols)
            self.np_vars = np.zeros(self._np_block_shape)
            self.np_t = np.zeros((self._np_block_rows,))
            self.np_epoch_t = np.zeros((self._np_block_rows,))
            self.np_k = np.zeros((self._np_block_rows,))

        # adding data to the matrix
        # self.np_vars[self.np_nrows, 0] = t
        self.np_t[self.np_nrows] = t
        self.np_epoch_t[self.np_nrows] = epoch_time
        self.np_k[self.np_nrows] = step
        self.np_vars[self.np_nrows, :] = np.array(var_data).reshape((-1))
        self.np_nrows += 1

        # check if matrix extension is needed
        if self.np_nrows >= self.np_vars.shape[0]:
            self.np_vars = np.concatenate([self.np_vars, np.zeros(self._np_block_shape)], axis=0)
            self.np_t = np.concatenate([self.np_t, np.zeros((self._np_block_rows,))], axis=0)
            self.np_epoch_t = np.concatenate([self.np_epoch_t, np.zeros((self._np_block_rows,))], axis=0)
            self.np_k = np.concatenate([self.np_k, np.zeros((self._np_block_rows,))], axis=0)

        # remove the post-computed variables from the variable list
        if self.system.tds.config.compute_flows:
            self.system.dae.y = self.system.dae.y[:self.system.dae.m]

    def get_latest_data(self):
        """
        Get the latest data and simulation time

        Returns
        -------
        dict : a dictionary containing `t` and `vars`
        """
        return {'epoch_time': float(self._last_epoch_t), 't': float(self._last_t), 'var': self._last_vars}

    def show(self):
        """
        The representation of an Varout object

        :return: the full result matrix (for use with PyCharm viewer)
        :rtype: np.array
        """
        out = []

        for item in self.vars:
            out.append(list(item))

        return np.array(out)

    def concat_t_vars(self):
        """
        Concatenate ``self.t`` with ``self.vars`` and output a single matrix
        for data dump

        :return matrix: concatenated matrix with ``self.t`` as the 0-th column
        """
        logger.warning('This function is deprecated and replaced by `concat_t_vars_np`.')
        out = np.array([])

        if len(self.t) == 0:
            return out

        out = np.ndarray(shape=(0, self.vars[0].size[0] + 1))

        for t, var in zip(self.t, self.vars):
            line = [[t]]
            line[0].extend(list(var))
            out = np.append(out, line, axis=0)

        return out

    def concat_t_vars_np(self, vars_idx=None):
        """
        Concatenate `self.np_t` with `self.np_vars` and return a single matrix.
        The first column corresponds to time, and the rest of the matrix is the variables.

        Returns
        -------
        np.array : concatenated matrix

        """
        selected_np_vars = self.np_vars
        if vars_idx is not None:
            selected_np_vars = self.np_vars[:, vars_idx]

        return np.concatenate([self.np_t[:self.np_nrows].reshape((-1, 1)),
                               selected_np_vars[:self.np_nrows, :]], axis=1)

    def get_xy(self, yidx, xidx=0):
        """
        Return stored data for the given indices for plot

        :param yidx: the indices of the y-axis variables(1-indexing)
        :param xidx: the index of the x-axis variables
        :return: None
        """
        if not isinstance(xidx, int):
            logger.error('Argument xidx must be an integer. get_xy() cannot continue')

        if isinstance(yidx, int):
            yidx = [yidx]

        t_vars = self.concat_t_vars()

        xdata = t_vars[:, xidx]
        ydata = t_vars[:, yidx]

        return xdata.tolist(), ydata.transpose().tolist()

    def dump_np_vars(self, store_format='csv', delimiter=','):
        """
        Dump the TDS simulation data to files by calling subroutines `write_lst` and
        `write_np_dat`.

        Parameters
        -----------

        store_format : str
            dump format in `('csv', 'txt', 'hdf5')`

        delimiter : str
            delimiter for the `csv` and `txt` format

        Returns
        -------
        bool: success flag
        """

        ret = False

        if self.system.files.no_output is True:
            logger.debug('no_output is True, thus no TDS dump saved ')
            return True

        if self.write_lst() and self.write_np_dat(store_format=store_format, delimiter=delimiter):
            ret = True

        return ret

    def dump(self):
        """
        Dump the TDS results to the output `dat` file

        :return: succeed flag
        """
        logger.warning('This function is deprecated and replaced by `dump_np_vars`.')

        ret = False

        if self.system.files.no_output:
            # return ``True`` because it did not fail
            return True

        if self.write_lst() and self.write_dat():
            ret = True

        return ret

    def write_np_dat(self, store_format='csv', delimiter=',', fmt='%.18e'):
        """
        Write TDS data stored in `self.np_vars` to the output file

        Parameters
        ----------
        store_format : str
            dump format in ('csv', 'txt', 'hdf5')

        delimiter : str
            delimiter for the `csv` and `txt` format

        fmt : str
            output formatting template

        Returns
        -------
        bool : success flag

        """
        ret = False
        system = self.system

        # compute the total number of columns, excluding time
        if not system.Recorder.n:
            n_vars = system.dae.m + system.dae.n
            # post-computed power flows include:
            #   bus   - (Pi, Qi)
            #   line  - (Pij, Pji, Qij, Qji, Iij_Real, Iij_Imag, Iji_real, Iji_Imag)
            if system.tds.config.compute_flows:
                n_vars += 2 * system.Bus.n + 8 * system.Line.n + 2 * system.Area.n_combination
            idx = list(range(n_vars))

        else:
            n_vars = len(system.Recorder.varout_idx)
            idx = system.Recorder.varout_idx

        # prepare data
        t_vars_concatenated = self.concat_t_vars_np(vars_idx=idx)

        try:
            os.makedirs(os.path.abspath(os.path.dirname(system.files.dat)), exist_ok=True)
            with open(system.files.dat, self._mode) as f:
                if store_format in ('csv', 'txt'):
                    np.savetxt(f, t_vars_concatenated, fmt=fmt, delimiter=delimiter)
                elif store_format == 'hdf5':
                    pass
                ret = True
                logger.info('TDS data dumped to <{}>'.format(system.files.dat))

        except IOError:
            logger.error('I/O Error while writing the dat file.')

        return ret

    def write_dat(self):
        """
        Write ``system.Varout.vars`` to a ``.dat`` file
        :return:
        """
        logger.warn('This function is deprecated and replaced by `write_np_dat`.')

        ret = False
        system = self.system

        # compute the total number of columns, excluding time
        if not system.Recorder.n:
            n_vars = system.dae.m + system.dae.n
            # post-computed power flows include:
            #   bus   - (Pi, Qi)
            #   line  - (Pij, Pji, Qij, Qji, Iij_Real, Iij_Imag, Iji_real, Iji_Imag)
            if system.tds.config.compute_flows:
                n_vars += 2 * system.Bus.n + 8 * system.Line.n + 2 * system.Area.n_combination
            idx = list(range(n_vars))

        else:
            n_vars = len(system.Recorder.varout_idx)
            idx = system.Recorder.varout_idx

        template = ['{:<8g}'] + ['{:0.10f}'] * n_vars
        template = ' '.join(template)

        # format the output in a string
        out = ''
        for t, line in zip(self.t, self.vars):
            values = [t] + list(line[idx])
            out += template.format(*values) + '\n'

        try:
            os.makedirs(os.path.abspath(os.path.dirname(system.files.dat)), exist_ok=True)
            with open(system.files.dat, self._mode) as f:
                f.write(out)
            ret = True

        except IOError:
            logger.error('I/O Error while writing the dat file.')

        return ret

    def write_lst(self):
        """
        Dump the variable name lst file

        :return: succeed flag
        """

        ret = False
        out = ''
        system = self.system
        dae = self.system.dae
        varname = self.system.varname
        template = '{:>6g}, {:>25s}, {:>35s}\n'

        # header line
        out += template.format(0, 'Time [s]', '$Time\\ [s]$')

        # include line flow variables in algebraic variables
        nflows = 0
        if self.system.tds.config.compute_flows:
            nflows = 2 * self.system.Bus.n + \
                     8 * self.system.Line.n + \
                     2 * self.system.Area.n_combination

        # output variable indices
        if system.Recorder.n == 0:
            state_idx = list(range(dae.n))
            algeb_idx = list(range(dae.n, dae.n + dae.m + nflows))
            idx = state_idx + algeb_idx
        else:
            idx = system.Recorder.varout_idx

        # variable names concatenated
        uname = varname.unamex + varname.unamey
        fname = varname.fnamex + varname.fnamey

        for e, i in enumerate(idx):
            # `idx` in the lst file is always consecutive
            out += template.format(e + 1, uname[i], fname[i])

        try:
            with open(self.system.files.lst, 'w') as f:
                f.write(out)
            ret = True
        except IOError:
            logger.error('I/O Error while writing the lst file.')

        return ret

    def vars_to_array(self):
        """
        Convert `self.vars` to a numpy array

        Returns
        -------
        numpy.array
        """

        logger.warn('This function is deprecated. You can inspect `self.np_vars` directly as NumPy arrays '
                    'without conversion.')
        if not self.vars:
            return None

        vars_matrix = matrix(self.vars, size=(self.vars[0].size[0],
                                              len(self.vars))).trans()

        self.vars_array = np.array(vars_matrix)

        return self.vars_array
